"""
Provides the Parabola Class, mainly used for Voronoi calculation
"""
import logging as root_logger
import numpy as np
from .quadratic import Quadratic as Q

logging = root_logger.getLogger(__name__)
MAX     = 5000

@dataclass
class ParabolaVertex:
    """
    Vertex form of a parabola:
    y = a(x-h)^2 + k
    """
    a : float = field(default=0)
    h : float = field(default=0)
    k : float = field(default=0)

    def __call__(self, x):
        return self.a * pow(x + self.h, 2) + self.k

    def to_standard_form(self):
        """ Calculate the standard form of the parabola from a vertex form """
        return StandardForm(self.a, -2*self.a*self.h, a*pow(self.h, 2)+self.k)

@dataclass
class ParabolaStandard:
    """
    standard form of a parabolar:
    y = ax^2 + bx + c
    """
    a : float = field()
    b : float = field()
    c : float = field()

    def __str__(self):
        return "y = {0:.2f} * x^2 + {1:.2f} x + {2:.2f}".format(self.a, self.b, self.c)

    def __call__(self, x):
        return (self.a * pow(x, 2)) + (self.b * x) + self.c

    def to_vertex_form(self):
        """ Calculate the vertex form of the parabola from a standard form """
        return VertexForm(a, -b/(2*a), c-(a * (pow(b, 2) / 4 * a)))

@dataclass
class ParabolaData:
    """
    A class to represent and calculate a parabola. holds both forms of definition,
    focus/directrix AND standard form.
    Handles the degenerate case of focus and directrix being the same by designating
    as a vertical line
    """
    fx            : float = field()
    fy            : float = field()
    d             : float = field()
    vertical_line : bool  = field(default=True)
    p             : float = field(init=False, default=0)

    vert          : VertexForm   = field(init=False)
    stan          : StandardForm = field(init=False)

    id = 0

    def __post_init__(self):
        """ Create a parabola with a focus x and y, and a directrix d """
        #pylint: disable=invalid-name
        self.id            = Parabola.id
        Parabola.id += 1
        #focal parameter: the distance from vertex to focus/directrix
        self.p = 0.5 * (self.fy - self.d)

        va = 0
        if not np.allclose(self.fy, self.d):
            va = 1/(2*(self.fy-self.d))

        self.vert = VertexForm(va, -self.fx, self.fy - self.p)
        self.stan = StandardForm(va, 2 * va * -self.fx, va * (pow(-self.fx, 2)) + self.self.fy - self.p)

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return str(self.stan)

    def __repr__(self):
        return "<Parabola>"

    def __call__(self, x, maximum=MAX):
        """ Shorthand for calculating an the entire parabola """
        if self.vertical_line:
            ys = np.linspace(self.fy, maximum)
            xs = np.repeat(self.fx, len(ys))
            return np.column_stack((xs, ys))
        else:
            return np.column_stack((x, self.calc_standard_form(x)))

    def __eq__(self, parabola2):
        """ Compare two parabolas """
        if not isinstance(parabola2, (type(None), Parabola)):
            raise TypeError()

        if parabola2 is None:
            return False
        a = self.to_numpy_array()
        b = parabola2.to_numpy_array()
        return np.allclose(a, b)

class Parabola:
    """ A class to represent and calculate a parabola. holds both forms of definition,
    focus/directrix AND standard form.
    Handles the degenerate case of focus and directrix being the same by designating
    as a vertical line
    """
    id = 0

    def __init__(self, fx, fy, d):
        """ Create a parabola with a focus x and y, and a directrix d """
        #pylint: disable=invalid-name
        self.id = Parabola.id
        Parabola.id += 1
        self.vertical_line = True
        self.fx = fx
        self.fy = fy
        self.d = d
        #focal parameter: the distance from vertex to focus/directrix
        self.p = 0.5 * (self.fy - self.d)
        #Vertex form: y = a(x-h)^2 + k
        if np.allclose(self.fy, self.d):
            self.va = 0
        else:
            self.va = 1/(2*(self.fy-self.d))
        self.vh = -self.fx
        self.vk = self.fy - self.p
        #standard form: y = ax^2 + bx + c
        self.sa = self.va
        self.sb = 2 * self.sa * self.vh
        self.sc = self.sa * (pow(self.vh, 2)) + self.vk

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return "y = {0:.2f} * x^2 + {1:.2f} x + {2:.2f}".format(self.sa, self.sb, self.sc)

    def __repr__(self):
        return "y = {} * x^2 + {} x + {}".format(self.sa, self.sb, self.sc)

    def is_left_of_focus(self, x):
        """ Given an x, is it smaller than the focus of this parabola """
        return x < self.fx

    def update_d(self, d):
        """ Update the parabola given the directrix has moved """
        self.d = d
        self.p = 0.5 * (self.fy - self.d)
        #Vertex form parameters:
        if np.allclose(self.fy, self.d):
            self.va = 0
            self.vertical_line = True
        else:
            self.va = 1/(2*(self.fy-self.d))
            self.vertical_line = False
        self.vk = self.fy - self.p
        #standard form:
        self.sa = self.va
        self.sb = 2 * self.sa * self.vh
        self.sc = self.sa * (pow(self.vh, 2)) + self.vk

    def intersect(self, p2, d=None):
        """ Take the quadratic representations of parabolas, and
            get the 0, 1 or 2 points that are the intersections
        """
        assert(isinstance(p2, Parabola))
        if d:
            self.update_d(d)
            p2.update_d(d)
        #degenerate cases:
        if self.vertical_line:
            logging.debug("Intersecting vertical line")
            return np.array([p2(self.fx)[0]])
        if p2.vertical_line:
            logging.debug("Intersecting other vertical line")
            return np.array([self(p2.fx)[0]])
        #normal (as quadratic class):
        q1 = Q(self.sa, self.sb, self.sc)
        q2 = Q(p2.sa, p2.sb, p2.sc)
        xs = q1.intersect(q2)
        #logging.debug("Resulting intersects: {}".format(xs))
        xys = self(xs)
        return xys

    def calc_standard_form(self, x):
        """ Get the y value of the parabola at an x position using the standard
            form equation. Should equal calc_vertex_form(x)
        """
        return (self.sa * pow(x, 2)) + (self.sb * x) + self.sc

    def calc_vertex_form(self, x):
        """ Get the y value of the parabola at an x position using
            the vertex form equation. Should equal calc_standard_form(x)
        """
        return self.va * pow(x + self.vh, 2) + self.vk

    def calc(self, x):
        """ For given xs, return an (n, 2) array of xy pairs of the parabola """
        return np.column_stack((x, self.calc_vertex_form(x)))

    def __call__(self, x, maximum=MAX):
        """ Shorthand for calculating an the entire parabola """
        if self.vertical_line:
            ys = np.linspace(self.fy, maximum)
            xs = np.repeat(self.fx, len(ys))
            return np.column_stack((xs, ys))
        else:
            return np.column_stack((x, self.calc_standard_form(x)))

    def to_numpy_array(self):
        """ Converts the parabola to a 1d array """
        return np.array([
            self.fx, self.fy,
            self.va, self.vh, self.vk,
            self.sa, self.sb, self.sc
            ])

    def to_array(self):
        """ Utility method to call to_numpy_array """
        return self.to_numpy_array()

    def __eq__(self, parabola2):
        """ Compare two parabolas """
        assert(isinstance(parabola2, Parabola))
        if parabola2 is None:
            return False
        a = self.to_numpy_array()
        b = parabola2.to_numpy_array()
        return np.allclose(a, b)

    def get_focus(self):
        """ Get the Focus point of this parabola. 2d np.array """
        return np.array([self.fx, self.fy])

    @staticmethod
    def to_standard_form(a, h, k):
        """ Calculate the standard form of the parabola from a vertex form """
        return [
            a,
            -2*a*h,
            a*pow(h, 2)+k
        ]

    @staticmethod
    def to_vertex_form(a, b, c):
        """ Calculate the vertex form of the parabola from a standard form """
        return [
            a,
            -b/(2*a),
            c-(a * (pow(b, 2) / 4 * a))
        ]



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import pathlib as pl
import textwrap
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
##-- end imports

@dataclass
class TwitterUsersWriter:
    """
    Utility class to write twitter users to a simple table file
    """
    _path : pl.Path

    CHARWIDTH        = 80

    @property
    def path(self):
        return self._path.expanduser().resolve()

    def start(self):
        if self.path.exists():
            return

        with open(self.path, 'a') as f:
            f.write("| User ID | Username | Tags | Verified |  Description|\n")
            f.write("|----|\n")

    def add(self, data):
        with open(self.path, 'a') as f:
            for user_str, username, verified, description in data:
                safe_desc = textwrap.wrap(description.replace('|',''), self.CHARWIDTH)
                if not bool(safe_desc):
                    safe_desc = [description]
                f.write("| {} | {} |  | {} | {} |\n".format(user_str,
                                                            username,
                                                            verified,
                                                            safe_desc[0]))
                for subline in safe_desc[1:]:
                    f.write("| | | | | {} |\n".format(subline))
                f.write ("|-----|\n")



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

def check_distance_from_points(point, quad_tree, dist=HALFDELTA):
    """ Given a point and a quadtree, return true if the point is within
    the bounds of the quadtree but not near any points """
    bbox = [point[0] - dist, point[1] - dist, point[0] + dist, point[1] + dist]
    area = quad_tree.intersect(bbox)
    inside_canvas = point[0] > 0 and point[0] < 1.0 and point[1] > 0 and point[1] < 1.0
    return (not bool(area)) and inside_canvas

def get_closest_to_focus(focus, possible_points):
    """ Given a set of points, return the point closest to the focus """
    ds = get_distance(focus, possible_points)
    m_d = ds.min()
    i = ds.tolist().index(m_d)
    return possible_points[i]

def get_closest_on_side(ref_point, possible_points, left=True):
    """
    given a reference point and a set of candidates, get the closest
    point on either the left or right of that reference
    """
    subbed = possible_points - ref_point
    if left:
        on_side = subbed[:, 0] < 0
    else:
        on_side = subbed[:, 0] > 0
    try:
        i = on_side.tolist().index(True)
        return possible_points[i]
    except ValueError:
        return None

def get_distance_raw(p1, p2):
    """ Get the non-square-root distance for pairs of points """
    assert(isinstance(p1, np.ndarray))
    assert(isinstance(p2, np.ndarray))
    p1 = p1.reshape(-1, 2)
    p2 = p2.reshape(-1, 2)
    d_squared = pow(p2-p1, 2)
    #summed = dSquared[:, 0] + dSquared[:, 1]
    summed = d_squared.sum(axis=1)
    return summed

def get_distance(p1, p2):
    """ Get the square-root distance of pairs of points """
    assert(isinstance(p1, np.ndarray))
    assert(isinstance(p2, np.ndarray))

    summed = get_distance_raw(p1, p2)
    sqrtd = np.sqrt(summed)
    return sqrtd

def get_distance_xyxy(x1, y1, x2, y2):
    """ Utility to get the raw distance of points as separate x's and y's  """
    return get_distance_raw(np.array([x1, y1]), np.array([x2, y2]))[0]

def get_midpoint(p1, p2):
    """ Given two points, get the point directly between them """
    m = (p1 + p2) / 2
    return m

def radians_between_points(a, b):
    """
    takes np.arrays
    return the radian relation of b to a (source)
    ie: if > 0: anti-clockwise, < 0: clockwise
    """
    c = b - a
    return atan2(c[1], c[0])




#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

printer = logmod.getLogger("doot._printer")

from time import sleep
import sh
import shutil
import tomlguard as TG
import jgdv

##-- expansion keys
TO_KEY             : Final[DootKey] = DootKey.make("to")
FROM_KEY           : Final[DootKey] = DootKey.make("from")
UPDATE             : Final[DootKey] = DootKey.make("update_")
PROMPT             : Final[DootKey] = DootKey.make("prompt")
PATTERN            : Final[DootKey] = DootKey.make("pattern")
SEP                : Final[DootKey] = DootKey.make("sep")
TYPE_KEY           : Final[DootKey] = DootKey.make("type")
AS_BYTES           : Final[DootKey] = DootKey.make("as_bytes")
FILE_TARGET        : Final[DootKey] = DootKey.make("file")
RECURSIVE          : Final[DootKey] = DootKey.make("recursive")
LAX                : Final[DootKey] = DootKey.make("lax")
##-- end expansion keys

COMP_TAR_CMD  = sh.tar.bake("-cf", "-")
COMP_GZIP_CMD = sh.gzip.bake("--best")
DECOMP_CMD    = sh.tar.bake("-xf")

@jgdv.check_protocol
class TarCompressAction(Action_p):
    """ Compresses a target into a .tar.gz file """

    @DootKey.kwrap.paths("file")
    @DootKey.kwrap.paths("to", hint={"on_fail":None})
    def __call__(self, spec, state, file, to):
        target = file
        output = to or target.with_suffix(target.suffix + ".tar.gz")

        if output.exists():
            raise doot.errors.DootActionError("Compression target already exists")
        if target.is_dir():
            COMP_GZIP_CMD(_in=COMP_TAR_CMD("-C", target, ".", _piped=True), _out=output)
        else:
            COMP_GZIP_CMD(_in=COMP_TAR_CMD("-C", target.parent, target.name, _piped=True), _out=output)

@jgdv.check_protocol
class TarDecompressAction(Action_p):
    """ Decompresses a .tar.gz file """

    @DootKey.kwrap.paths("file", "to")
    def __call__(self, spec, state, file, to):
        target = file
        output = to
        if not ".tar.gz" in target.name:
            printer.warning("Decompression target isn't a .tar.gz", target)
            return ActionResponseEnum.FAIL

        DECOMP_CMD(target, "-C", output)


@jgdv.check_protocol
class TarListAction(Action_p):
    """ List the contents of a tar archive """

    @DootKey.kwrap.paths("from")
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, _from, _update):
        target = _from
        if "".join(target.suffixes) != ".tar.gz":
            printer.warning("Trying to list the contents of a non-tar archive")
            return ActionResponseEnum.FAIL

        result = sh.tar("--list", "-f", str(target))
        lines = result.split("\n")
        return { _update : lines }





"""
Defines the PatternConstructor, used in parsing

PatternConstructor is a contextual class (__enter__ and __exit__)
that values are added to.

Upon exit, it calculates the appropriate time arcs for events

"""
##-- imports
from __future__ import annotations

from .arc import Arc
from .event import Event
from .pattern import Pattern
from collections import namedtuple
from enum import Enum
from fractions import Fraction as t
import logging as root_logger
##-- end imports

logging = root_logger.getLogger(__name__)

# PStart/End : SubPattern
# PDual : layering
# CStart/End : Choice
# OP : Optional
# SIL : Silence
CTOR_ACT = Enum("Actions for Pattern Constructor",
                "PSTART PEND PDUAL CSTART CEND OP SIL")

#Placeholders:
PatternPH = namedtuple("PatternPH", "values")
ChoicePH = namedtuple("ChoicePH", "values")
ParallelPH = namedtuple("ParallelPH", "values")

def construct_pattern(tokens):
    """   """
    #init stack with a top level pattern placeholder
    stack = []
    balance_count = 0
    while bool(tokens):
        logging.debug("Tokens: {}".format(tokens))
        head = tokens.pop(0)
        logging.debug("Head: {}".format(head))
        logging.debug("Stack: {}".format(stack))
        if isinstance(head, CTOR_ACT):
            if any(head is x for x in [CTOR_ACT.PSTART,
                                       CTOR_ACT.CSTART]):
                balance_count += 1
                logging.debug("Starting a Pattern or Choice")
                stack.append([])
            elif head is CTOR_ACT.PEND:
                logging.debug("Pattern end")
                balance_count -= 1
                p_vals = stack.pop()
                if bool(stack) and isinstance(stack[-1], ParallelPH):
                    logging.debug("Part of a Parallel Pattern")
                    parPH = stack.pop()
                    parPH.values.append(p_vals)
                    p_vals = parPH

                if isinstance(p_vals, ParallelPH):
                    #is parallel, do for each in parallel
                    sub_vals = [prepare_pvals(x) for x in p_vals.values]
                    processed = ParallelPH(sub_vals)
                else:
                    processed = prepare_pvals(p_vals)

                if bool(stack):
                    stack[-1].append(processed)
                else:
                    stack.append(processed)
            elif head is CTOR_ACT.PDUAL:
                logging.debug("Parallel")
                #run patterns in parallel
                #get the current pattern
                curr = stack.pop()
                if not bool(stack) or not isinstance(stack[-1], ParallelPH):
                    logging.debug("Creating new ParallelPH")
                    newPH = ParallelPH([curr])
                    stack.append(newPH)
                    stack.append([])
                else:
                    logging.debug("Using Existing ParallelPH")
                    stack[-1].values.append(curr)
                    stack.append([])
            elif head is CTOR_ACT.CEND:
                logging.debug("Choice End")
                balance_count -= 1
                c_vals = stack.pop()
                placeholder = ChoicePH(c_vals)
                stack[-1].append(placeholder)
            elif head is CTOR_ACT.OP:
                logging.warning("OP not implemented yet")

            elif head is CTOR_ACT.SIL:
                stack[-1].append(head)
            else:
                logging.warning("Other: {}".format(head))
        else:
            #push to list on the top of the stack
            stack[-1].append(head)
        logging.debug("Post Stack: {}".format(stack))
        logging.debug("------")

    if (balance_count != 0):
        raise Exception("Brackets in patterns need to be balanced")
    assert(len(stack) == 1)
    #create the final pattern
    farc= Arc(t(0,1),t(1,1))
    finalPH = stack.pop()
    if isinstance(finalPH, ParallelPH):
        finalEvents = [Event(farc, Pattern(farc, x.values), True) for x in finalPH.values]
    elif isinstance(finalPH, list):
        finalEvents = [Event(farc, Pattern(farc, x.values), True) for x in finalPH]
    else:
        finalEvents = finalPH.values

    result= Pattern(farc, finalEvents)
    return result

def prepare_pvals(p_vals):
    """
    Prepare pattern values for assembling into a pattern
    """
    logging.debug("Preparing: %s", p_vals)
    p_len     = len(p_vals)
    time_vals = []
    #calc times
    for i,x in enumerate(p_vals):
        arc = Arc(t(i,p_len), t(i+1, p_len))
        val = None
        if isinstance(x, PatternPH):
            val = [Event(arc, Pattern(arc, x.values), True)]
        elif isinstance(x, ParallelPH):
            pats = [Pattern(arc,y.values) for y in x.values]
            val = [Event(arc, y, True) for y in pats]
        elif isinstance(x, ChoicePH):
            #TODO: Make a Choice Event
            logging.warning("Choice events aren't implemented yet")
        else:
            val = [Event(arc, x)]
        time_vals += val
    #create pattern placeholder
    placeholder = PatternPH(time_vals)
    #add to new top of stack
    logging.debug("Resulting Placeholder: ", placeholder)
    return placeholder



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import pathlib as pl
import datetime
import json
import logging as logmod
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
from uuid import uuid1

import networkx as nx
##-- end imports

logging = logmod.getLogger(__name__)

@dataclass
class TwitterGraph:
    """
    A Directed Graph of tweets, built from a directory of tweet jsons
    """
    graph : nx.DiGraph = field(default_factory=nx.DiGraph)

    def add_tweet(self, tweet:dict, source:pl.Path):
        """
        Add tweet data to the graph
        reply_id -> tweet_id so threads are root -> reply -> reply -> reply -> leaf
        tweet_id -> quoted_id so quotes are tweet -> quote -> quote...
        """
        tweet_id = tweet.get('id_str', None)
        user_id  = tweet.get('user', {}).get('id_str', None)

        if tweet_id is None:
            return

        if tweet_id not in self.graph:
            self.graph.add_node(tweet_id)

        self.graph.nodes[tweet_id].update({"source_file":str(source), "user":user_id})

        match tweet:
            case {'in_reply_to_status_id_str': None, 'is_quote_status': False}:
                pass
            case { 'in_reply_to_status_id_str': str() as reply_id, 'in_reply_to_user_id_str': reply_user, 'is_quote_status': False }:
                self.graph.add_node(reply_id, user=reply_user, source_file=str(source))
                self.graph.add_edge(reply_id, tweet_id, type="reply")
            case { 'in_reply_to_status_id_str': str() as reply_id, 'in_reply_to_user_id_str': reply_user, 'quoted_status_id_str': str() as quoted_id }:
                self.graph.add_node(reply_id, user=reply_user, source_file=str(source))
                quoted_user = tweet.get('quoted_status', {}).get('user', {}).get('id_str', None)
                if quoted_user is not None:
                    self.graph.add_node(quoted_id, user=quoted_user, source_file=str(source))

                self.graph.add_edge(reply_id, tweet_id, type="reply")
                self.graph.add_edge(tweet_id, quoted_id, type="quote")
            case { 'quoted_status_id_str': str() as quoted_id }:
                quoted_user = tweet.get('quoted_status', {}).get('user', {}).get('id_str', None)
                if quoted_user is not None:
                    self.graph.add_node(quoted_id, user=quoted_user, source_file=str(source))

                self.graph.add_edge(tweet_id, quoted_id, type="quote")

    def components(self):
        logging.info("Getting Components for Graph")
        undir_graph : nx.Graph = nx.Graph([x for x in self.graph.edges if self.graph.get_edge_data(*x)['type'] == 'reply'])
        components             = list(nx.connected_components(undir_graph))
        logging.info("Components Retrieved, getting quotes")
        found = set()
        for component in components:
            quotes = self.get_quotes(*component)
            component.update(quotes)
            found.update(component)

        rest = set(undir_graph.nodes) - found

        return components, rest

    def get_quotes(self, *srcs: str):
        if not bool(srcs):
            return [ y for x,y in self.graph.edges if self.graph.get_edge_data(x,y).get('type', "n/a") == 'quote']

        queue      = [x for id_s in srcs for x in self.graph[id_s] if self.graph.get_edge_data(id_s, x)['type'] == "quote"]
        discovered = set()
        while bool(queue):
            quoted_id = queue.pop(0)
            if quoted_id in discovered:
                continue
            discovered.add(quoted_id)
            queue += [x for x in self.graph[quoted_id] if self.graph.get_edge_data(quoted_id, x)['type'] == "quote"]

        return discovered

    def roots(self):
        reply_graph : nx.DiGraph = self.graph.edge_subgraph([x for x in self.graph.edges if self.graph.get_edge_data(*x).get('type', "n/a") == 'reply'])
        return [x for x,y in reply_graph.in_degree if y == 0]

    def reply_chains(self, roots) -> list[list[str]]:
        reply_graph : nx.DiGraph = self.graph.edge_subgraph([x for x in self.graph.edges if self.graph.get_edge_data(*x).get('type', "n/a") == 'reply'])

        discovered = set()
        chains     = []
        current    = []
        last       = []
        for root in roots:
            for l, r, kind in nx.dfs_labeled_edges(reply_graph, root):
                discovered.add(r)
                match kind:
                    case 'forward':
                        current.append(r)
                    case 'reverse' if not all(x in last for x in current):
                        chains.append(current[:])
                        current.pop()
                        last = current[:]
                    case 'reverse':
                        current.pop()

        remaining = [x for x in set(reply_graph.nodes) - discovered]
        return chains, remaining



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

import pyparsing as pp
logging = logmod.getLogger(__name__)
ModuleFragment = AT.ModuleFragment

@register_class("print")
class PrintCmd:
    """
    Print out details of the current state.

    Usage:
    print wm                : prints the current working memory
    print (ctx|c)           : prints the current active context instances
    print ctx[{num}]        : prints a specific context instance by index
    print ctx[{num}:{num}]  : prints a slice of the context instances
    print ctx {binding*}    : prints a specific binding from the context instances

    """

    def __init__(self):
        self._parser = self._gen_parser()

    def _gen_parser(self):
        wm_kw        = pp.Keyword("wm")("wm")
        mod_target   = pp.Optional(pp.Word(pp.alphas + ".")("mod_target"))
        module_kw    = pp.MatchFirst([pp.Keyword("mod"),
                                      pp.Keyword("module")])("module")
        semantic_kw  = pp.Keyword("semantics")("semantics")

        # TODO also handle bind without kw?
        printer_parser = pp.MatchFirst([wm_kw,
                                        module_kw + mod_target,
                                        semantic_kw,
                                        ctx_parser,
                                        rst])

        return printer_parser

    def __call__(self, line):
        try:
            params = self._parser.parse_string(line, parse_all=True)
        except pp.ParseException as err:
            print("Bad Print Command: ", err.pstring)
            print(self.__doc__)
            return

        logging.info(f"Printing: {line}")
        result = []
        if "wm" in params:
            # TODO print everything from a query down
            print(self._repl.state.engine.pprint())
        elif "context" in params or "short_context" in params or "context_slice" in params:
            self.prep_print_contexts(params)

    def prep_print_contexts(self, params):
        ctxs_to_print     = []
        bindings_to_print = []

        # Get the contexts to print
        if "short_context" in params:
            try:
                ctxs_to_print.append(self._repl.state.ctxs[params['short_context']])
            except IndexError as err:
                print(f"Selected bad ctx instance. Try 0 <= x < {len(self._repl.state.ctxs)}.")

        elif "context_slice" in params:
            ctx_slice = self._repl.state.ctxs[params['context_slice']].active_list()
            ctxs_to_print += ctx_slice
        elif bool(self._repl.state.ctxs) and len(self._repl.state.ctxs) > 0:
            ctxs_to_print += self._repl.state.ctxs.active_list()
        elif bool(self._repl.state.ctxs._named_sets):
            print("Named (continuation) Sets:")
            print(self._repl.state.engine.pprint(target=list(self._repl.state.ctxs._named_sets.keys())))

        if not bool(ctxs_to_print):
            print(f"No applicable contexts to print")
            return

        if "bindings" in params:
            bindings_to_print += params.bindings[:]

        self.print_contexts(params, ctxs_to_print, bindings_to_print)

    def print_contexts(self, params, ctxs_to_print, bindings):
        logging.info("Ctxs: {}", ctxs_to_print)
        logging.info("Bindings: {}", bindings)
        # now print them
        for i,ctx in enumerate(ctxs_to_print):
            if len(ctxs_to_print) > 1:
                print(f"Context: {i}")
            else:
                print("Context: ")
            # if bool(ctx.continuation):
            #     print(f"Continuation: {ctx.continuation}")
            if bool(bindings):
                for x in bindings:
                    print("{:<5} : {}".format(x, self._repl.state.engine.pprint(target=[ctx[x]])))
            else:
                for x,y in ctx.data.items():
                    print("{:<5} : {}".format(x, self._repl.state.engine.pprint(target=[y])))

            print("--------------------")



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import importlib
from tomlguard import TomlGuard

PAD           : Final[int] = 15
TaskFlagNames : Final[str] = [x.name for x in TaskFlags]

def aware_splitter(x, sep="."):
    match x:
        case str():
            return x.split(sep)
        case _:
            return [x]

@dataclass(eq=False, slots=True)
class StructuredName:
    """ A Complex name class for identifying tasks and classes.

      Classes are the standard form used in importlib: "module.path:ClassName"
      Tasks use a double colon to separate head from tail name: "group.name::TaskName"

    """
    head            : list[str]          = field(default_factory=list)
    tail            : list[str|UUID]     = field(default_factory=list)

    separator       : str                = field(default=doot.constants.patterns.TASK_SEP, kw_only=True)
    subseparator    : str                = field(default=".", kw_only=True)

    def __post_init__(self):
        sub_split = ftz.partial(aware_splitter, sep=self.subseparator)
        match self.head:
            case None | []:
                self.head = ["default"]
            case ["tasks", x] if x.startswith('"') and x.endswith('"'):
                self.head = ftz.reduce(lambda x, y: x + y, map(sub_split, x[1:-1]))
            case ["tasks", *xs]:
                self.head = ftz.reduce(lambda x, y: x + y, map(sub_split, xs))
            case list():
                self.head = ftz.reduce(lambda x, y: x + y, map(sub_split, self.head))
            case str():
                self.head = self.head.split(self.subseparator)
            case _:
                self.head = [self.head]

        match self.tail:
            case None | []:
                self.tail = ["default"]
            case list():
                self.tail = ftz.reduce(lambda x, y: x + y, map(sub_split, self.tail))
            case str():
                self.tail = self.tail.split(self.subseparator)
            case _:
                self.tail = [self.tail]

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return self.head_str + self.separator + self.tail_str

    def __lt__(self, other) -> bool:
        """ Compare two names, return true if other is a subname of this name
        eg: a.b.c < a.b.c.d
        """
        match other:
            case str():
                other = StructuredName.build(other)
            case StructuredName():
                pass
            case _:
                return False

        for x,y in zip(self.head, other.head):
            if x != y:
                return False

        for x,y in zip(self.tail, other.tail):
            if x != y:
                return False

        return True

    def __contains__(self, other:str):
        return str(other) in str(self)

    def __eq__(self, other):
        return str(self) == str(other)

    def tail_str(self):
        return self.subseparator.join(str(x) for x in self.tail)

    def head_str(self):
        return self.subseparator.join(str(x) for x in self.head)



# pylint: disable=bad-whitespace
"""

Defines functions for parsers and parse -> data transform

"""
##-- imports
from __future__ import annotations

import logging as logmod

import pyparsing as pp

##-- end imports

logging = logmod.getLogger(__name__)

ATOM : str = TYPE_BASE

def strip_parse_type(s, loc, toks):
    """ Utility function to strip out parse data from return tuples,
    useful for:
    [("QUERY", actual_query)] -> [actual_query]

    NOTE: expects to be called from a group wrapping the actual parser
    """
    assert(all([isinstance(x, tuple) for x in toks[0]]))
    return [x[1] for x in toks[0]]

def deep_update_names(parser):
    logging.debug("Deep Updating Parser Names for {}", parser)
    queue = [parser]
    processed = set()

    while bool(queue):
        current = queue.pop(0)
        if current in processed:
            continue
        processed.add(current)

        if hasattr(current, "_defaultName"):
            setattr(current, "_defaultName", None)

        if hasattr(current, "expr"):
            queue.append(current.expr)
        elif hasattr(current, "exprs"):
            queue += current.exprs

def clear_parser_names(*parsers):
    logging.debug("Clearing Parser Names for: {}", parsers)
    for parser in parsers:
        # if hasattr(parser, "customName"):
        #     setattr(parser, "customName", None)

        if hasattr(parser, "_defaultName"):
            setattr(parser, "_defaultName", None)



""" The highest level data structure in a dcel, apart from the dcel itself """
##-- imports
from __future__ import annotations
import logging as root_logger
from dataclasses import InitVar, dataclass, field
from itertools import cycle, islice
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import numpy as np
from scipy.spatial import ConvexHull

from ..maths import umath as cumath
from ..constants import END, FACE, SMALL_RADIUS, START, TWOPI, VERTEX, WIDTH
from ..drawing import clear_canvas, draw_circle, draw_text
from ..umath import calc_bbox_corner, within_bbox
from .constants import EditE, FaceE
from .drawable import Drawable
from .halfedge import HalfEdge
from .vertex import Vertex

##-- end imports

logging = root_logger.getLogger(__name__)

@dataclass
class Face(Drawable):
    """
    A Face with a start point for its outer component list,
    and all of its inner components
    """

    #Site is the voronoi point that the face is built around
    site               : np.ndarray     = field()
    dcel               : DCEL           = field()
    #Primary list of ccw edges for this face
    edge_list          : List[Any]      = field(default_factory=list)
    coord_list         : List[Any]      = field(default_factory=list)
    #free vertices to build a convex hull from:
    free_vertices      : Set[Any]       = field(default_factory=set)
    data               : Dict[Any, Any] = field(default_factory=dict)
    index              : int            = field(default=None)

    marked_for_cleanup : bool           = field(init=False, default=False)
    nextIndex = 0

    def __post_init__(self):
        if site is not None:
            #site = np.array([0, 0])
            assert(isinstance(site, np.ndarray))

        if self.index is None:
            logging.debug("Creating Face {}".format(Face.nextIndex))
            self.index = Face.nextIndex
            Face.nextIndex += 1
        else:
            assert(isinstance(index, int))
            logging.debug("Re-creating Face: {}".format(index))
            self.index = index
            if self.index >= Face.nextIndex:
                Face.nextIndex = self.index + 1

        if self.dcel is not None and self not in self.dcel.faces:
            self.dcel.faces.add(self)


    def copy(self):
        """" Create a copy of the entire face """
        with self.dcel:
            #copy the halfedges
            es = [x.copy() for x in self.edge_list]
            #create a new face
            f = self.dcel.new_face(edges=es)
            #copy the data
            f.data.update(self.data)
            #return it
            return f

    #------------------------------
    # def hulls
    #------------------------------

    @staticmethod
    def hull_from_vertices(verts):
        """ Given a set of vertices, return the convex hull they form,
        and the vertices to discard """
        #TODO: put this into dcel?
        assert(all([isinstance(x, Vertex) for x in verts]))
        #convert to numpy:
        np_pairs = [(x.to_array(), x) for x in verts]
        hull = ConvexHull([x[0] for x in np_pairs])
        hull_verts = [np_pairs[x][1] for x in hull.vertices]
        discard_verts = set(verts).difference(hull_verts)
        assert(not bool(discard_verts.intersection(hull_verts)))
        assert(len(discard_verts) + len(hull_verts) == len(verts))
        return (hull_verts, list(discard_verts))

    @staticmethod
    def hull_from_coords(coords):
        """ Given a set of coordinates, return the hull they would form
        DOESN NOT RETURN DISCARDED, as the coords are not vertices yet
        """
        assert(isinstance(coords, np.ndarray))
        assert(coords.shape[1] == 2)
        hull = ConvexHull(coords)
        hull_coords = np.array([coords[x] for x in hull.vertices])
        return hull_coords


    #------------------------------
    # def Human Readable Representations
    #------------------------------


    def __str__(self):
        return "Face: {}".format(self.get_centroid())

    def __repr__(self):
        edge_list = len(self.edge_list)
        return "<Face: {}, edge_list: {}>".format(self.index, edge_list)

    def draw(self, ctx, clear=False, force_centre=False, text=False, data_override=None):
        """ Draw a single Face from a dcel.
        Can be the only thing drawn (clear=True),
        Can be drawn in the centre of the context for debugging (force_centre=True)
        """
        data = self.data.copy()
        if data_override is not None:
            assert(isinstance(data, dict))
            data.update(data_override)

        #early exits:
        if len(self.edge_list) < 2:
            return
        #Custom Clear
        if clear:
            clear_canvas(ctx)

        #Data Retrieval:
        line_wdith         = WIDTH
        vert_colour        = START
        vert_rad           = SMALL_RADIUS
        face_col           = FACE
        radius             = SMALL_RADIUS
        text_string        = "F: {}".format(self.index)
        should_offset_text = FaceE.TEXT_OFFSET in data
        centroid_col       = VERTEX
        draw_centroid      = FaceE.CENTROID in data
        sample_descr       = None

        if draw_centroid and isinstance(data[FaceE.CENTROID], (list, np.ndarray)):
            centroid_col = data[FaceE.CENTROID]
        if FaceE.STARTVERT in data and isinstance(data[FaceE.STARTVERT], (list, np.ndarray)):
            vert_colour = data[FaceE.STARTVERT]
        if FaceE.STARTRAD in data:
            vert_rad = data[FaceE.STARTRAD]
        if FaceE.FILL in data and isinstance(data[FaceE.FILL], (list, np.ndarray)):
            face_col = data[FaceE.FILL]
        if FaceE.CEN_RADIUS in data:
            radius = data[FaceE.CEN_RADIUS]
        if FaceE.TEXT in data:
            text_string = data[FaceE.TEXT]
        if FaceE.WIDTH in data:
            line_wdith = data[FaceE.WIDTH]
        if FaceE.SAMPLE in data:
            sample_descr = data[FaceE.SAMPLE]


        #Centre to context
        mid_point = (self.dcel.bbox[2:] - self.dcel.bbox[:2]) * 0.5
        face_centre = self.get_centroid()
        if force_centre:
            inv_centre = -face_centre
            ctx.translate(*inv_centre)
            ctx.translate(*mid_point)

        if sample_descr is not None:
            #draw as a sampled line
            sample_descr(ctx, self)

        if FaceE.NULL in data:
            return

        ctx.set_line_width(line_wdith)
        ctx.set_source_rgba(*face_col)
        #Setup Edges:
        initial = True
        for x in self.get_edges():
            v1, v2 = x.get_vertices()
            assert(v1 is not None)
            assert(v2 is not None)
            logging.debug("Drawing Face {} edge {}".format(self.index, x.index))
            logging.debug("Drawing Face edge from ({}, {}) to ({}, {})".format(v1.loc[0],
                                                                               v1.loc[1],
                                                                               v2.loc[0],
                                                                               v2.loc[1]))
            if initial:
                ctx.move_to(*v1.loc)
                initial = False
            ctx.line_to(*v2.loc)

            #todo move this out
            if FaceE.STARTVERT in data:
                ctx.set_source_rgba(*vert_colour)
                draw_circle(ctx, *v1.loc, vert_rad)


        #****Draw*****
        if FaceE.FILL not in data:
            ctx.stroke()
        else:
            ctx.close_path()
            ctx.fill()


        #Drawing the Centroid point
        ctx.set_source_rgba(*END)
        if draw_centroid:
            ctx.set_source_rgba(*centroid_col)
            draw_circle(ctx, *face_centre, radius)

        #Text Retrieval and drawing
        if text or FaceE.TEXT in data:
            draw_text(ctx, *face_centre, text_string, offset=should_offset_text)

        #Reset the forced centre
        if force_centre:
            ctx.translate(*(mid_point * -1))


    #------------------------------
    # def Exporting
    #------------------------------


    def _export(self):
        """ Export identifiers rather than objects, to allow reconstruction """
        logging.debug("Exporting face: {}".format(self.index))
        enum_data     = {a.name:b for a, b in self.data.items() if a in FaceE}
        non_enum_data = {a:b for a, b in self.data.items() if a not in FaceE}

        return {
            'i'             : self.index,
            'edges'         : [x.index for x in self.edge_list if x is not None],
            'sitex'         : self.site[0],
            'sitey'         : self.site[1],
            "enum_data"     : enum_data,
            "non_enum_data" : non_enum_data
        }


    def get_bbox(self):
        """ Get a rough bbox of the face """
        #TODO: fix this? its rough
        vertices      = [x.origin for x in self.edge_list]
        vertex_arrays = [x.to_array() for x in vertices if x is not None]
        if not bool(vertex_arrays):
            return np.array([[0, 0], [0, 0]])
        all_vertices = np.array([x for x in vertex_arrays])
        bbox         = np.array([np.min(all_vertices, axis=0),
                                 np.max(all_vertices, axis=0)])
        logging.debug("Bbox for Face {}  : {}".format(self.index, bbox))
        return bbox

    def mark_for_cleanup(self):
        """ Schedules the face for cleanup """
        self.marked_for_cleanup = True


    #------------------------------
    # def centroids
    #------------------------------

    def get_centroid(self):
        """ Get the user defined 'centre' of the face """
        if self.site is not None:
            return self.site.copy()
        else:
            return self.get_avg_centroid()

    def get_avg_centroid(self):
        """ Get the averaged centre point of the face from the vertices of the edges """
        k = len(self.edge_list)
        coords = np.array([x.origin.loc for x in self.edge_list])
        norm_coord = np.sum(coords, axis=0) / k
        if self.site is None:
            self.site = norm_coord
        return norm_coord

    def get_centroid_from_bbox(self):
        """ Alternate Centroid, the centre point of the bbox for the face"""
        bbox = self.get_bbox()
        difference = bbox[1, :] - bbox[0, :]
        centre = bbox[0, :] + (difference * 0.5)
        if self.site is None:
            self.site = centre
        return centre


    #------------------------------
    # def edge access
    #------------------------------

    def get_edges(self):
        """ Return a copy of the edgelist for this face """
        return self.edge_list.copy()

    def add_edges(self, edges):
        """ Add a list of edges to the face """
        assert(isinstance(edges, list))
        for x in edges:
            self.add_edge(x)

    def add_edge(self, edge):
        """ Add a constructed edge to the face """
        assert(isinstance(edge, HalfEdge))
        if edge.face is self:
            return
        if edge.face is not self and edge.face is not None:
            edge.face.remove_edge(edge)
        self.coord_list = None
        edge.face = self
        if edge not in self.edge_list:
            self.edge_list.append(edge)
        edge.marked_for_cleanup = False

    def remove_edge(self, edge):
        """ Remove an edge from this face, if the edge has this face
        registered, remove that too """
        assert(isinstance(edge, HalfEdge))
        #todo: should the edge be connecting next to prev here?
        if not bool(self.edge_list):
            return
        if edge in self.edge_list:
            self.edge_list.remove(edge)
        if edge.face is self:
            edge.face = None
        if edge.twin is None or edge.twin.face is None:
            edge.mark_for_cleanup()

    def sort_edges(self):
        """ Order the edges clockwise, by starting point, ie: graham scan """
        logging.debug("Sorting edges")
        centre = self.get_centroid()
        #verify all edges are ccw
        edges = self.edge_list.copy()
        for x in edges:
            if not x.he_ccw(centre):
                x.swapFaces()

        assert(all([x.he_ccw(centre) for x in self.edge_list]))
        # withDegrees = [(x.degrees(centre), x) for x in self.edge_list]
        # withDegrees.sort()
        # self.edge_list = [hedge for (deg, hedge) in withDegrees]
        self.edge_list.sort()

    def has_edges(self):
        """ Check if its a null face or has actual edges """
        return bool(self.edge_list)


    #------------------------------
    # def modifiers
    #------------------------------

    def subdivide(self, edge, ratio=None, angle=0):
        """ Bisect / Divide a face in half by creating a new line
        on the ratio point of the edge, at the angle specified, until it intersects
        a different line of the face.
        Angle is +- from 90 degrees.
        returns the new face
        """
        self.sort_edges()
        if ratio is None:
            ratio = 0.5
        assert(isinstance(edge, HalfEdge))
        assert(edge in self.edge_list)
        assert(0 <= ratio <= 1)
        assert(-90 <= angle <= 90)
        #split the edge
        new_point, new_edge = edge.split_by_ratio(ratio)

        #get the bisecting vector
        as_coords = edge.to_array()
        bisector = cumath.get_bisector(as_coords[0], as_coords[1])
        #get the coords of an extended line
        extended_end = cumath.extend_line(new_point.to_array(), bisector, 1000)
        el_coords = np.row_stack((new_point.to_array(), extended_end))

        #intersect with coords of edges
        intersection = None
        opp_edge = None
        for he in self.edge_list:
            if he in [edge, new_edge]:
                continue
            he_coords = he.to_array()
            intersection = cumath.intersect(el_coords, he_coords)
            if intersection is not None:
                opp_edge = he
                break
        assert(intersection is not None)
        assert(opp_edge is not None)
        #split that line at the intersection
        new_opp_point, new_opp_edge = opp_edge.split(intersection)

        #create the other face
        new_face = self.dcel.new_face()

        #create the subdividing edge:
        dividing_edge = self.dcel.new_edge(new_point, new_opp_point,
                                           face=self,
                                           twin_face=new_face,
                                           edata=edge.data,
                                           vdata=edge.origin.data)
        dividing_edge.add_prev(edge, force=True)
        dividing_edge.add_next(new_opp_edge, force=True)
        dividing_edge.twin.add_prev(opp_edge, force=True)
        dividing_edge.twin.add_next(new_edge, force=True)

        #divide the edges into new_opp_edge -> edge, new_edge -> opp_edge
        new_face_edge_group = []
        original_face_edge_update = []

        current = new_opp_edge
        while current != edge:
            assert(current.next is not None)
            original_face_edge_update.append(current)
            current = current.next
        original_face_edge_update.append(current)
        original_face_edge_update.append(dividing_edge)

        current = new_edge
        while current != opp_edge:
            assert(current.next is not None)
            new_face_edge_group.append(current)
            current.face = new_face
            current = current.next
        new_face_edge_group.append(current)
        current.face = new_face
        new_face_edge_group.append(dividing_edge.twin)

        #update the two faces edgelists
        self.edge_list = original_face_edge_update
        new_face.edge_list = new_face_edge_group

        #return both
        return (self, new_face)

    @staticmethod
    def merge_faces(*args):
        """ Calculate a convex hull from all passed in faces,
        creating a new face """
        assert(all([isinstance(x, Face) for x in args]))
        dc = args[0].dcel
        assert(dc is not None)
        all_verts = set()
        for f in args:
            all_verts.update(f.get_all_vertices())
        new_face = dc.new_face()
        #then build the convex hull
        hull, discarded = Face.hull_from_vertices(all_verts)
        for s, e in zip(hull, islice(cycle(hull), 1, None)):
            #create an edge
            dc.new_edge(s, e, face=new_face)
        #link the edges
        dc.link_edges_together(new_face.edge_list, loop=True)
        #return the face
        return (new_face, discarded)

    def translate_edge(self, transform, e=None, i=None, candidates=None, force=False):
        """ move an edge of the face """
        assert(e is None or e in self.edge_list)
        assert(i is None or 0 <= i < len(self.edge_list))
        assert(not (e is None and i is None))
        assert(isinstance(transform, np.ndarray))
        assert(transform.shape == (2, ))
        if i is None:
            i = self.edge_list.index(e)

        if not force and self.has_constraints(candidates):
            copied, _ = self.copy().translate_edge(transform, i=i, force=True)
            return (copied, EditE.NEW)

        self.edge_list[i].translate(transform, force=True)
        return (self, EditE.MODIFIED)

    def scale(self, amnt=None, target=None, vert_weights=None, edge_weights=None,
              force=False, candidates=None):
        """ Scale an entire face by amnt,
        or scale by vertex/edge normal weights """
        if not force and self.has_constraints(candidates):
            face_prime, _ = self.copy().scale(amnt=amnt, target=target,
                                              vert_weights=vert_weights,
                                              edge_weights=edge_weights,
                                              force=True)
            return (face_prime, EditE.NEW)

        if target is None:
            target = self.get_centroid_from_bbox()
        if amnt is None:
            amnt = np.ndarray([1, 1])
        assert(isinstance(amnt, np.ndarray))
        assert(amnt.shape == (2, ))
        if vert_weights is not None:
            assert(isinstance(vert_weights, np.ndarray))
        if edge_weights is not None:
            assert(isinstance(edge_weights, np.ndarray))

        verts = self.get_all_vertices()
        for vert in verts:
            loc = vert.loc.copy()
            loc -= target
            loc *= amnt
            loc += target
            vert.translate(loc, absolute=True, force=True)

        return (self, EditE.MODIFIED)

    def cut_out(self, candidates=None, force=False):
        """ Cut the Face out from its verts and halfedges that comprise it,
        creating new verts and edges, so the face can be moved and scaled
        without breaking the already existing structure """
        if not force and self.has_constraints(candidates):
            return (self.copy(), EditE.NEW)
        else:
            return (self, EditE.MODIFIED)

    def rotate(self, rads, target=None, candidates=None, force=False):
        """ copy and rotate the entire face by rotating each point """
        assert(-TWOPI <= rads <= TWOPI)
        if not force and self.has_constraints(candidates):
            face_prime, _ = self.copy().rotate(rads, target=target, force=True)
            return (face_prime, EditE.NEW)

        if target is None:
            target = self.get_centroid_from_bbox()
        assert(isinstance(target, np.ndarray))
        assert(target.shape == (2, ))

        for l in self.edge_list:
            l.rotate(c=target, r=rads, candidates=candidates, force=True)
        return (self, EditE.MODIFIED)

    def constrain_to_circle(self, centre, radius, candidates=None, force=False):
        """ Constrain the vertices and edges of a face to be within a circle """
        if not force and self.has_constraints(candidates):
            logging.debug("Face: Constraining a copy")
            face_prime, _ = self.copy().constrain_to_circle(centre, radius, force=True)
            return (face_prime, EditE.NEW)

        logging.debug("Face: Constraining edges")
        #constrain each edge
        edges = self.edge_list.copy()
        for e in edges:
            logging.debug("HE: {}".format(e))
            eprime, edit_e = e.constrain_to_circle(centre, radius, force=True)
            logging.debug("Result: {}".format(eprime))
            assert(edit_e == EditE.MODIFIED)
            assert(eprime in self.edge_list)
            if eprime.marked_for_cleanup:
                self.edge_list.remove(eprime)

        return (self, EditE.MODIFIED)

    #todo: possibly add a shrink/expand to circle method
    def constrain_to_bbox(self, bbox, candidates=None, force=False):
        """ Given a bbox, ensure all edges of the face are within """
        if not force and self.has_constraints(candidates):
            face_prime, _ = self.copy().constrain_to_bbox(bbox, force=True)
            return (face_prime, EditE.NEW)

        edges = self.edge_list.copy()

        for edge in edges:
            if edge.outside(bbox):
                self.remove_edge(edge)
                continue
            edge.constrain_to_bbox(bbox, candidates=candidates, force=True)

        return (self, EditE.MODIFIED)


    #------------------------------
    # def Vertex access
    #------------------------------

    def add_vertex(self, vert):
        """ Add a vertex, then recalculate the convex hull """
        assert(isinstance(vert, Vertex))
        self.free_vertices.add(vert)
        self.coord_list = None

    def get_all_vertices(self):
        """ Get all vertices of the face. both free and in halfedges """
        all_verts = set()
        all_verts.update(self.free_vertices)
        for e in self.edge_list:
            all_verts.update(e.get_vertices())
        return all_verts

    def get_all_coords(self):
        """ Get the sequence of coordinates for the edges """
        if self.coord_list is not None:
            return self.coord_list
        all_coords = np.array([x.to_array() for x in self.get_all_vertices()])
        self.coord_list = Face.hull_from_coords(all_coords)
        return self.coord_list


    #------------------------------
    # def verification
    #------------------------------

    def fixup(self, bbox=None):
        """ Verify and enforce correct designations of
        edge ordering, next/prev settings, and face settings """
        assert(bbox is not None)
        if not bool(self.edge_list):
            self.mark_for_cleanup()
            return []
        if len(self.edge_list) < 2:
            return []

        for e in self.edge_list:
            self.add_edge(e)

        altered = False
        centre = self.get_centroid()
        avg_centre = self.get_avg_centroid()
        if not within_bbox(centre, bbox) and within_bbox(avg_centre, bbox):
            altered = True
            self.site = avg_centre

        self.sort_edges()

        inferred_edges = []
        edges = self.edge_list.copy()
        prev = edges[-1]
        for e in edges:
            #enforce next and prev
            if e.prev is not prev:
                e.add_prev(prev, force=True)
            #if verts don't align AND they intersect the border of the bbox on separate edges:
            dont_align = not prev.connections_align(e)
            if dont_align:
                logging.debug("connections don't align")
                new_edge = self.dcel.new_edge(e.prev.twin.origin,
                                              e.origin,
                                              edata=e.data,
                                              vdata=e.origin.data)
                new_edge.face = self
                new_edge.add_prev(e.prev, force=True)
                new_edge.add_next(e, force=True)
                #insert that new edge into the edge_list
                index = self.edge_list.index(e)
                self.edge_list.insert(index, new_edge)
                inferred_edges.append(new_edge)

                #if the new_edge connects two different sides, split it and
                #force the middle vertex to the corner
                nib = new_edge.intersects_bbox(bbox)
                edge_es = set([ev for (coord, ev) in nib])
                if len(nib) == 2 and len(edge_es) > 1:
                    new_point, new_edge2 = new_edge.split_by_ratio(0.5, face_update=False)
                    new_edge2.face = self
                    self.edge_list.insert(index+1, new_edge2)
                    # if new_edge.twin.face is not None:
                    #     new_edge.twin.face.add_edge(new_edge2.twin)

                    move_to_coord = calc_bbox_corner(bbox, edge_es)
                    new_point.translate(move_to_coord, absolute=True, force=True)

            prev = e

        self.sort_edges()

        if altered:
            self.site = centre

        return inferred_edges

    def has_constraints(self, candidate_set=None):
        """ Tests whether the face's component edges and vertices are claimed by
        anything other than the face's own halfedges and their twins, and any passed in
        candidates """
        if candidate_set is None:
            candidate_set = set()
        candidates_plus_self = candidate_set.union([self],
                                                   self.edge_list,
                                                   [x.twin for x in self.edge_list
                                                    if x.twin is not None])
        return any([x.has_constraints(candidates_plus_self) for x in self.edge_list])

    def are_points_within(self, points):
        """ TODO: check whether a set of vertices are within the faces boundaries """
        assert(isinstance(points, np.ndarray))
        #see https://stackoverflow.com/questions/217578
        raise Exception("Unimplemented: are_points_within")



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging
import imageio

import cairo_utils as cu

img_types = cu.config.on_fail([".jpg"], list).tool.cairo.exts()

@dataclass
class GifData:
    """ A Utility class to easily create gifs from a number of images """

    output_dir  : str = field(default=".")
    gif_name    : str = field(default="anim.gif")
    source_dir  : str = field(default="images")
    file_format : str = field(default=".png")
    fps         : int = field(default=12)

    num_regex = re.compile(r'(\d+)')

class MakeGif:

    def get_num(self, text) -> int:
        """ Given a String, extract a number from it,
        or return a default """
        logging.info("Getting num of: %s", text)
        assert(isinstance(text, str))
        try:
            return int(self.num_regex.search(s).group(0))
        except ValueException:
            return 9999999

    def make_gif(self, fpath:pl.Path):
        """ Trigger the creation of the GIF """
        # Get all Files
        frames = [x for x in fpath.iterdir() if x.suffix in img_types]
        assert(bool(frames))
        logging.info("Making gif of %s frames", len(frames))

        # Export as a Gif
        logging.info("Starting GIF writing")
        with imageio.get_writer(self.output, mode='I') as writer:
            # Sort by the number extracted from the filename
            for pframe in sorted(files, key=self.get_num):
                image = imageio.imread(pframe))
                writer.append_data(image)

        logging.info("Finished GIF writing")



"""
Utils related to rational time
"""
##-- imports
from __future__ import annotations
from enum import Enum
from fractions import Fraction, gcd
from functools import reduce
import logging as root_logger

##-- end imports

import datetime

logging   = root_logger.getLogger(__name__)

PATTERN_T = Enum("Pattern Type", "DISCRETE ANALOG")
TIME_T    = Enum("Time Type", "CLOCK EVENT SET SYMBOLIC")

def time_str(time):
    return f"{time.numerator}/{time.denominator}"

def run_pattern(pat, num_cycles=3, action=None):
    #TODO
    raise NotImplementedError()

def print_run_pattern(pat, num_cycles=3):
    """
    Given a pattern, output the Events Per Cycle
    """
    #Calculate the subdivision necessary
    base_count = reduce(gcd, pat.base(), 2).denominator
    logging.info("Base Count: {}".format(base_count))
    #Track the active events
    curr_events = set()
    #Run a number of cycles:
    for x in range(num_cycles):
        logging.info("Cycle: {}".format(x))
        #Run the Cycle:
        for y in range(base_count):
            pos = (x * base_count) + y
            #Query the Pattern:
            q = pat(Time(pos, base_count))

            #Calculate the Event entries and exits
            turn_on = [x for x in q if x not in curr_events]
            turn_off = [x for x in curr_events if x not in q]
            #Update the active events set
            curr_events = set(q)

            #Output:
            turn_on_s = ", ".join([w.print_flip(True) for w in turn_on])
            turn_off_s = ", ".join([w.print_flip(False) for w in turn_off])
            logging.info("A: {} R: {} Events: {} {}".format(pos,
                                                            y,
                                                            turn_on_s,
                                                            turn_off_s))
        logging.info("----")

def roundTime(dt=None, roundTo=60):
   """Round a datetime object to any time lapse in seconds
   dt : datetime.datetime object, default now.
   roundTo : Closest number of seconds to round to, default 1 minute.
   Author: Thierry Husson 2012 - Use it as you want but don't blame me.
   from: https://stackoverflow.com/questions/3463930
   """
   dt       = dt or datetime.datetime.now()
   seconds  = (dt.replace(tzinfo=None) - dt.min).seconds
   rounding = (seconds+roundTo/2) // roundTo * roundTo
   rounded  = dt + datetime.timedelta(0,rounding-seconds,-dt.microsecond)
   return rounded



"""
Arcs express segments of rational time
"""
##-- imports
from __future__ import annotations

from fractions import Fraction
from .utils import time_str
import logging as root_logger
##-- end imports

logging = root_logger.getLogger(__name__)

class Arc:
    """ A segment of rational time """

    @staticmethod
    def copy(val):
        return Arc(val.start, val.end)

    def __init__(self, a, b):
        assert(isinstance(a, Fraction))
        assert(isinstance(b, Fraction))
        self.start = a
        self.end = b

    def __contains__(self, other):
        """ Test whether the given time is within bounds """
        test = other
        if not isinstance(other, Fraction):
            assert(hasattr(other, 'arc'))
            test = other.arc.start
        assert(isinstance(test, Fraction))
        return self.start <= test and test < self.end

    def __repr__(self):
        """ A Readable format of an arc """
        return f"({time_str(self.start)}..{time_str(self.end)})"

    def __eq__(self, other):
        if not isinstance(other, Arc):
            raise TypeError()

        return all([x == y for x,y in zip(self.pair(), other.pair())])

    def pair(self):
        """ Treat the arc as a list """
        return [self.start, self.end]

    def size(self):
        """ Get the length of time the arc describes """
        return self.end - self.start

    def bound(self, other):
        assert(isinstance(other, Arc))
        start = min(self.start, other.start)
        end = max(self.end, other.end)
        return Arc(start, end)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

def granulate(xys, grains=10, mult=2, random=None):
    """ Given a set of points, offset each slightly
    by the direction between the points
    """
    assert(isinstance(xys, np.ndarray))
    assert(len(xys.shape) == 2)
    if random is None:
        random = np.random.random
    directions, hypos = get_directions(xys)
    granulated = np.zeros((1, 2))
    for i, d in enumerate(hypos):
        sub_granules = xys[i, :] + (d * directions[i, :]*(random((grains, 1))) * mult)
        granulated = np.row_stack((granulated, sub_granules))
    return granulated[1:]

def vary(xys, step_size, pix, random=None):
    """
    FIXME : investigate
    for a given set of points, wiggle them slightly
    """
    assert(isinstance(xys, np.ndarray))
    assert(len(xys.shape) == 2)
    if random is None:
        random = np.random.random
    r = (1.0-2.0 * random((len(xys), 1)))
    scale = np.reshape(np.arange(len(xys)).astype('float'), (len(xys), 1))
    noise = (r* scale * step_size)
    a = random(len(xys))
    rnd = np.column_stack((np.cos(a), np.sin(a)))
    rnd_noise = rnd * noise
    rnd_noise_pix = rnd_noise * pix
    xys_prime = xys + rnd_noise_pix
    return xys_prime


def sample_wrapper(func, data, n, radius, colour, easing=None, random=None):
    """ A Wrapper for sample_circle, sample_along_lines,
    bezier1cp and bezier2cp
    Takes the function, applies the data and samp_sig_or_count to it,
    then duplicates the radius and colours to the appropriate count,
    and combines
    """
    assert(callable(func))
    sampled = func(data, n, easing=easing, random=random)
    radius_size = len(data) * n
    repeated_radius = np.repeat(radius, radius_size).reshape((radius_size, 1))
    repeated_colours = np.repeat(colour.reshape((-1,4)), len(sampled), axis=0)
    combined = np.column_stack((sampled, repeated_radius, repeated_colours))
    return combined


def displace_around_circle(xys, scale, n, random=None):
    """ displace the data around a scaled noisy circle """
    #pylint: disable=invalid-name
    #Create a circle:
    if random is None:
        random = np.random.random
    t = np.linspace(0, 2*pi, n)
    rotation = np.column_stack((sin(t), cos(t))).transpose()
    #create some noise:
    noise = random(n)
    #jitter the rotation:
    jittered = (rotation * noise)
    #control the amount of this noise to apply
    scaled = (jittered * scale).transpose()
    #apply the noise to the data
    mod_points = xys + scaled
    return mod_points


def calculate_single_point(points, d=DELTA, random=None):
    """ points passed in, move in a random direction """
    if random is None:
        random = np.random.random
    arr = random((points.shape[0], 2)) * TWOPI
    delta = np.array([sin(arr[:, 0]), cos(arr[:, 1])]) * (2 * d)
    return points + delta

def calculate_vector_point(ps, d=DELTA, random=None):
    """ passed in pairs of points, move in the direction of the vector """
    if random is None:
        random = np.random.random
    vector = ps[:, 2:] - ps[:, :2]
    rand_amnt = random((ps.shape[0], 2)) * TWOPI
    mag = np.sqrt(np.sum(np.square(vector)))
    norm_vector = vector / mag
    move_vector = norm_vector * (2 * d)
    jiggled_vector = move_vector * np.array([sin(rand_amnt[:, 0]), cos(rand_amnt[:, 1])])
    return ps[:, 2:] + jiggled_vector



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import contextlib
import io

# with contextlib.redirect_stdout(io.StringIO()) as stdout:
#      $0


class ColourStripPrintCapture(logmod.Formatter):
    """
      WARNING: MODIFIES builtins.print
      intercept print commands, so they can be logged at DEBUG level
    Force Colour Command codes to be stripped out of a string.
    """

    _default_fmt      = "{asctime} | {shortname:25} | {message}"
    _default_date_fmt = "%Y-%m-%d %H:%M:%S"
    _default_style    = '{'
    _colour_strip     = re.compile(r'\x1b\[([\d;]+)m?')

    def __init__(self, *, fmt=None, record=False):
        """
        Create a formatter given *Brace* style log format
        """
        super().__init__(fmt or self._default_fmt,
                         datefmt=self._default_date_fmt,
                         style=self._default_style)

    def format(self, record):
        result    = super().format(record)
        no_colour = self._colour_strip.sub("", result)
        return no_colour

    @staticmethod
    def capture_printing_to_file(path:str|pl.Path="print.log", *, disable_warning=False):
        """
        Setup a file handler for a separate logger,
        to keep a trace of anything printed.
        Strips colour print command codes out of any string
        printed strings are logged at DEBUG level
        """
        if not disable_warning:
            import warnings
            warnings.warn("Modifying builtins.print", RuntimeWarning)

        import builtins
        oldprint = builtins.print
        file_handler = logmod.FileHandler(path, mode='w')
        file_handler.setLevel(logmod.DEBUG)
        file_handler.setFormatter(ColourStripPrintCapture())

        print_logger = logmod.getLogger('print.intercept')
        print_logger.setLevel(logmod.NOTSET)
        print_logger.addHandler(file_handler)
        print_logger.propagate = False

        @wraps(oldprint)
        def intercepted(*args, **kwargs):
            """ Wraps `print` to also log to a separate file """
            oldprint(*args, **kwargs)
            if bool(args):
                print_logger.debug(args[0])

        builtins.print = intercepted

    @staticmethod
    def redirect_printing_to_logging(*, disable_warning=False):
        """ redirect printing into logging the logging system to handle
          logged at DEBUG level
        """
        if not disable_warning:
            import warnings
            warnings.warn("Modifying builtins.print", RuntimeWarning)

        import builtins
        oldprint = builtins.print
        print_logger = logmod.getLogger('print.intercept')
        print_logger.setLevel(logmod.DEBUG)

        @wraps(oldprint)
        def intercepted(*args, **kwargs):
            """ Wraps `print` to also log to a separate file """
            oldprint(*args, **kwargs)
            if bool(args):
                print_logger.debug(args[0])

        builtins.print = intercepted



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class DCELSubtractor:

    def reset_frontier(self):
        """ Clear the general incremental algorithm frontier """
        self.frontier = set([])

    def clear_quad_tree(self):
        """ Clear the internal quadtree of vertices """
        self.vertex_quad_tree = pyqtree.Index(bbox=self.bbox)
    def purge_edge(self, target):
        """ Clean up and delete an edge """
        assert(isinstance(target, HalfEdge))
        logging.debug("Purging Edge: {}".format(target.index))
        target_update = set()

        target.connect_next_to_prev()
        vert = target.origin
        target.origin = None
        if vert is not None:
            vert.unregister_half_edge(target)
            if vert.is_edgeless():
                vert.mark_for_cleanup()
                target_update.add(vert)

        if target.face is not None:
            face = target.face
            face.remove_edge(target)
            if not face.has_edges():
                face.mark_for_cleanup()
                target_update.add(face)

        if target.twin is not None:
            target.twin.mark_for_cleanup()
            target_update.add(target.twin)
            target.twin.twin = None
            target.twin = None


        self.half_edges.remove(target)

        return target_update

    def purge_vertex(self, target):
        """ Clean up and delete a vertex """
        assert(isinstance(target, Vertex))
        logging.debug("Purging Vertex: {}".format(target.index))
        target_update = set()

        half_edges = target.half_edges.copy()
        for edge in half_edges:
            assert(edge.origin == target)
            edge.origin = None
            target.unregister_half_edge(edge)
            edge.mark_for_cleanup()
            target_update.add(edge)

        self.vertices.remove(target)

        return target_update

    def purge_face(self, target):
        """ Clean up and delete a face """
        assert(isinstance(target, Face))
        logging.debug("Purging Face: {}".format(target.index))
        target_update = set()
        edges = target.get_edges()
        for edge in edges:
            target.remove_edge(edge)
            edge.mark_for_cleanup()
            target_update.add(edge)
        self.faces.remove(target)
        return target_update

    def purge(self, targets=None):
        """ Run all purge methods in correct order """
        if targets is None:
            #populate the targets:
            targets = set([])
            targets = targets.union([x for x in self.vertices if x.marked_for_cleanup])
            targets = targets.union([x for x in self.half_edges if x.marked_for_cleanup
                                     or x.is_infinite()])
            targets = targets.union([x for x in self.faces if x.marked_for_cleanup
                                     or not x.has_edges()])

        else:
            targets = set(targets)

        purged = set()
        while bool(targets):
            current = targets.pop()
            if current in purged:
                continue
            if isinstance(current, Vertex):
                targets = targets.union(self.purge_vertex(current))
            elif isinstance(current, HalfEdge):
                targets = targets.union(self.purge_edge(current))
            elif isinstance(current, Face):
                targets = targets.union(self.purge_face(current))
            purged.add(current)

        self.calculate_quad_tree()


    #------------------------------
    # def Vertex, Edge, HalfEdge Creation
    #------------------------------



""" Vertex: The lowest level data structure in a dcel """
##-- imports
from __future__ import annotations

import logging as root_logger
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (ALLCLOSE_TOLERANCE, D_EPSILON, TOLERANCE,
                                   TWOPI, VERTEX, VERTRAD, EditE, VertE)
from cairo_utils.drawable import Drawable
from cairo_utils.drawing import draw_circle
from cairo_utils.umath import in_circle, rotate_point

##-- end imports

logging = root_logger.getLogger(__name__)

@dataclass
class Vertex(Drawable):
    """ A Simple vertex for two dimensions.
    Has a pair of coordinates, and stores the edges associated with it.
    """

    loc                               : np.ndarray     = field()
    dcel                              : DCEL           = field()
    #The edges this vertex is part of :
    half_edges                        : List[HalfEdge] = field(default_factory=list)
    #Custom data of the vertex        :
    data                              : Dict[Any, Any] = field(default_factory=dict)
    index                             : int            = field(default=None)

    marked_for_cleanup                : bool           = field(init=False, default=False)
    active                            : bool           = field(init=False, default=True)
    nextIndex = 0

    def __post_init__(self):
        assert(isinstance(loc, np.ndarray))
        assert(edges is None or isinstance(edges, list))

        if index is None:
            logging.debug("Creating vertex {} at: {:.3f} {:.3f}".format(Vertex.nextIndex,
                                                                        loc[0], loc[1]))
            self.index = Vertex.nextIndex
            Vertex.nextIndex += 1
        else:
            assert(isinstance(index, int))
            logging.debug("Re-Creating Vertex: {}".format(index))
            self.index = index
            if self.index >= Vertex.nextIndex:
                Vertex.nextIndex = self.index + 1

        if self.dcel is not None and self not in self.dcel.vertices:
            self.dcel.vertices.add(self)
            self.dcel.vertex_quad_tree.insert(item=self, bbox=self.bbox())



    def __repr__(self):
        edges = [x.index for x in self.half_edges]
        edges += [x.twin.index for x in self.half_edges]
        return "<Vertex V: {}, edges: {}, ({:.3f}, {:.3f})>".format(self.index, edges,
                                                                    self.loc[0], self.loc[1])

    def __lt__(self, other):
        """ Sorting top to bottom, left to right """
        assert(isinstance(other, Vertex))
        if np.allclose(self.loc[1], other.loc[1],
                       atol=ALLCLOSE_TOLERANCE[0],
                       rtol=ALLCLOSE_TOLERANCE[1]):
            return self.loc[0] < other.loc[0]
        else:
            return self.loc[1] > other.loc[1]

    def copy(self):
        """ Create an isolated copy of this vertex. Doesn't copy halfedge connections,
        but does copy data """
        new_vert = self.dcel.new_vertex(self.loc, data=self.data.copy())
        return new_vert

    def mark_for_cleanup(self):
        """ Schedule this object to be deleted on the next pass """
        self.marked_for_cleanup = True

    #------------------------------
    # def exporting
    #------------------------------
    def _export(self):
        """ Export identifiers instead of objects to allow reconstruction """
        logging.debug("Exporting Vertex: {}".format(self.index))
        enum_data     = {a.name:b for a, b in self.data.items() if a in VertE}
        non_enum_data = {a:b for a, b in self.data.items() if a not in VertE}

        return {
            'i'             : self.index,
            'x'             : self.loc[0],
            'y'             : self.loc[1],
            'half_edges'    : [x.index for x in self.half_edges],
            "enum_data"     : enum_data,
            "non_enum_data" : non_enum_data,
            "active"        : self.active
        }


    #------------------------------
    # def Human Readable Representations
    #------------------------------

    # def __str__(self):
    #     return "({:.3f}, {:.3f})".format(self.loc[0], self.loc[1])


    #------------------------------
    # def activation
    #------------------------------
    def activate(self):
        """ Activates the object, so it is drawn """
        self.active = True

    def deactivate(self):
        """ Removes the object from being drawn """
        self.active = False


    #------------------------------
    # def bboxes
    #------------------------------

    def bbox(self, e=D_EPSILON):
        """ Create a minimal bbox for the vertex,
        for dcel to find overlapping vertices using a quadtree  """
        return Vertex.free_bbox(self.loc, e=e)

    @staticmethod
    def free_bbox(loc, e=D_EPSILON):
        """ Static method utility to create a bbox.
        used for quad_tree checking without creating the vertex """
        assert(isinstance(loc, np.ndarray))
        loc = loc.astype(np.float64)
        return np.array([loc - e, loc + e]).flatten()


    #------------------------------
    #def queries
    #------------------------------
    def is_edgeless(self):
        """ asks whether a vertext has no registered halfedges  """
        return not bool(self.half_edges)

    def has_constraints(self, candidate_set=None):
        """ if a vertex is used by more than  """
        if candidate_set is None:
            candidate_set = set()
        assert(isinstance(candidate_set, set))
        return bool(self.half_edges.difference(candidate_set))

    def get_nearby_vertices(self, e=D_EPSILON):
        """ Utility method to get nearby vertices through the dcel reference "",
        returns the list of matches *including* self """
        assert(self.dcel is not None)
        return self.dcel.vertex_quad_tree.intersect(self.bbox(e=e))

    def within(self, bbox, tolerance=TOLERANCE):
        """ Check the vertex is within [x, y, x2, y2] """
        assert(isinstance(bbox, np.ndarray))
        assert(len(bbox) == 4)
        mod_bbox = bbox + np.array([-tolerance, -tolerance, tolerance, tolerance])
        in_x_bounds = mod_bbox[0] < self.loc[0] and self.loc[0] < mod_bbox[2]
        in_y_bounds = mod_bbox[1] < self.loc[1] and self.loc[1] < mod_bbox[3]
        return in_x_bounds and in_y_bounds

    def within_circle(self, centre, radius):
        """ Check the vertex is within the radius boundary of a point """
        return in_circle(centre, radius, self.to_array())[0]

    def outside(self, bbox):
        """ Check the vertex is entirely outside of the bbox [x, y, x2, y2] """
        return not self.within(bbox)


    #------------------------------
    # def HalfEdge Access and Registration
    #------------------------------

    def register_half_edge(self, he):
        """ register a halfedge as using this vertex
        will add the vertex into the first open slot of the halfedge
        """
        #Don't assert isinstance, as that would require importing halfedge
        assert(hasattr(he, 'index'))
        self.half_edges.add(he)
        logging.debug("Registered v{} to e{}".format(self.index, he.index))

    def unregister_half_edge(self, he):
        """ Remove a halfedge from the list that uses this vertex,
        also removes the vertex from the half_edges' slot
        """
        assert(hasattr(he, 'index'))
        if he in self.half_edges:
            self.half_edges.remove(he)
        logging.debug("Remaining edges: {}".format(len(self.half_edges)))

    def get_sorted_edges(self):
        """ return all half-edges that this vertex starts,
        sorted by angle. always relative to unit vector (right) """
        opp_hedges = {x.twin.origin : x for x in self.half_edges}
        verts = opp_hedges.keys()
        sorted_verts = self.dcel.order_vertices(self.loc, verts)
        return [opp_hedges[x] for x in sorted_verts]


    #------------------------------
    # def Coordinate access
    #------------------------------
    def to_array(self):
        """ Convert the Vertex's coords to a simple numpy array """
        return self.loc


    #------------------------------
    # Def Modifiers
    #------------------------------
    def extend_line_to(self, direction=None, length=None, rad=None, target=None, edge_data=None):
        """ create a line extending out from this vertex  """
        #TODO: calc target from direction, len, rad
        if target is None:
            raise Exception("Target is None")
        assert(isinstance(target, np.ndarray))
        assert(self.dcel is not None)
        new_edge = self.dcel.create_edge(self.to_array(),
                                        target,
                                        vdata=self.data,
                                        edata=edge_data)

        #make the edge have faces:
        self.register_half_edge(new_edge)
        return new_edge

    def translate(self, direction, d=1, absolute=False, candidates=None, force=False):
        """ Move the vertex by the vector dir, scaled by distance d,
        if absolute is true, just move to the specified point"""
        assert(isinstance(direction, np.ndarray))
        assert(direction.shape == (2, ))
        if not absolute:
            target = self.to_array() + (direction * d)
        else:
            target = direction
        if not force and self.has_constraints(candidates):
            return (self.dcel.new_vertex(target), EditE.NEW)
        else:
            self.loc = target
            return (self, EditE.MODIFIED)

    def rotate(self, c=None, r=0, candidates=None, force=False):
        """ Rotate the point around a target """
        assert(isinstance(c, np.ndarray))
        assert(c.shape == (2, ))
        assert(-TWOPI <= r <= TWOPI)
        new_loc = rotate_point(self.to_array(), cen=c, rads=r)
        if not force and self.has_constraints(candidates):
            return (self.dcel.new_vertex(new_loc), EditE.NEW)
        else:
            self.loc = new_loc
            return (self, EditE.MODIFIED)

    def draw(self, ctx, data_override=None):
        data = self.data.copy()
        if data_override is not None:
            data.update(data_override)

        vert_col = VERTEX
        vert_rad = VERTRAD
        sample_description = None

        if VertE.STROKE in data and isinstance(data[VertE.STROKE], (list, np.ndarray)):
            vert_col = data[VertE.STROKE]
        if VertE.RADIUS in data:
            vert_rad = data[VertE.RADIUS]
        if VertE.SAMPLE in data:
            sample_description = data[VertE.SAMPLE]

        if sample_description is not None:
            #draw as a sampled line
            sample_description(ctx, self)

        if VertE.NULL in data:
            return

        ctx.set_source_rgba(*vert_col)
        draw_circle(ctx, *self.loc, vert_rad)



"""

reminder: https://docs.python.org/3/library/cmd.html
"""
##-- imports
from __future__ import annotations

import cmd
import logging as logmod
import sys
import traceback
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

##-- end imports

##-- logging
logging      = logmod.getLogger(__name__)
trace_logger = logmod.getLogger('acab.repl.trace')
##-- end logging

import pyparsing as pp
from .repl_state import ReplState
from .error import JGDVREPLException

repl_intro = ["Welcome to DejaVu's Default Repl Commander.",
              "Type 'help' or '?' to list commands.",
              "Type 'tutorial' for a tutorial.", "Type ':q' to quit."
              ]

class REPLCommander(cmd.Cmd):
    """ Implementation of cmd.Cmd to provide an extensible REPL"""
    intro                                                   = "\n".join(repl_intro)
    prompt                                                  = initial_prompt + ": "
    _latebind                                               = []
    _default_startups  : ClassVar[list[Callable[..., Any]]] = []

    state  : ReplState = ReplState()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for inst in self._latebind:
            assert(getattr(inst, "_repl") is None)
            setattr(inst, "_repl", self)

        for fn in self._default_startups:
            fn(self, "")

    def default(self, line):
        """ Called when no other command matches """
        # default to assertion / query / run
        try:
            self.state.ctxs = self.state.engine(line,
                                                ctxset=self.state.ctxs)
        except JGDVREPLParseException as err:
            print(str(err))
        except JGDVREPLException as err:
            logging.warning("\n--------------------\nFailure:\n")
            traceback.print_tb(err.__traceback__)
            logging.warning(f"\n{err.args[-1]}\n")
            print(str(err))

    def precmd(self, line):
        """ For massaging the input command """
        # convert symbols -> cmd names.
        # eg: ':{' -> multi
        trace_logger.info("[repl]>>> " + line)
        try:
            logging.debug("PreCmd Parsing: {}".format(line))
            line = RP.precmd_parser.parse_string(line)[:]
            logging.debug("PreCmd Result:{}".format(line))
            # Intercept if in multi line state
            if self.state.in_multi_line and not line[0] in ["multi", "pop", "exit", "echo"]:
                logging.info("In Multi")
                line = ["collect"] + line

            if bool(self.state.echo):
                print(f"{line}")

            return " ".join(line)

        except pp.ParseException as err:
            traceback.print_tb(err.__traceback__)
            logging.warning(f"Parse Failure: {err.markInputline()}")

    def onecmd(self, line):
        try:
            return super(REPLCommander, self).onecmd(line)
        except Exception as err:
            traceback.print_tb(err.__traceback__)
            print(f"\n{err.args[-1]}\n")

    def postcmd(self, stop, line):
        """
        Update the repl prompt to display number of viable contexts
        """
        for name, hook in sorted(self.state.post_cmds.items()):
            hook(self)

        return stop

    def parseline(self, line):
        """Parse the line into a command name and a string containing
        the arguments.  Returns a tuple containing (command, args, line).
        'command' and 'args' may be None if the line couldn't be parsed.
        """
        if not self.state.in_multi_line:
            line = line.strip()

        if not line:
            return None, None, line
        elif line[0] == '?':
            line = 'help ' + line[1:]
        elif line[0] == '!':
            if hasattr(self, 'do_shell'):
                line = 'shell ' + line[1:]
            else:
                return None, None, line

        # split into cmd and args
        i, n = 0, len(line)
        while i < n and line[i] in self.identchars: i = i+1
        cmd, arg = line[:i], line[i:]

        if not self.state.in_multi_line:
            arg = arg.strip()

        return cmd, arg, line

    def emptyline(self):
        """ Overrides default of 'repeat last command',
        and prints the working memory
        """
        return self.onecmd("print wm")

    @classmethod
    def register(cls, fn):
        """ Decorator for registering a function into the repl """
        logging.debug(f"{cls.__name__} Registration: {fn.__name__}")
        assert("do_" in fn.__name__)
        assert(fn.__name__ not in dir(cls))
        setattr(cls, fn.__name__, fn)
        return fn

    @classmethod
    def register_class(cls, name):
        """ Register an entire class as the command bound to do_{name},
        (specifically the class' __call__ method)
        """

        def __register(target_cls):
            assert(hasattr(target_cls, "__call__"))
            assert(hasattr(cls, "_latebind"))
            if (not bool(target_cls.__call__.__doc__)) and bool(target_cls.__doc__):
                target_cls.__call__.__doc__ = target_cls.__doc__

            instance = target_cls()
            assert(not hasattr(target_cls, "_repl"))
            setattr(cls, f"do_{name}", instance.__call__)
            setattr(instance, "_repl", None)

            cls._latebind.append(instance)
            return target_cls

        return __register

    @classmethod
    def register_default(cls, fn):
        """
        Register and automatically call the function when REPLCommander is created.
        eg: register_default(do_ctxprompt) means the repl will show active context numbers
        from startup

        """
        assert(hasattr(cls, "_default_startups"))
        cls.register(fn)
        cls._default_startups.append(fn)



"""
Provides a Quadratic Clss, allowing manipulation of quadratic equations
"""
##-- imports
from __future__ import annotations
import logging as root_logger
from dataclasses import InitVar, dataclass, field
from math import sqrt
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import numpy as np

##-- end imports

logging = root_logger.getLogger(__name__)

@dataclass
class Quadratic:
    """ A Class to hold a quadratic equation
    (y = ax^2 + bx + c)
    and perform operations on it
    """

    a : float = field(default=0)
    b : float = field(default=0)
    c : float = field(default=0)

    def __repr__(self):
        return f"<Quadratic: {self.a} x^2 + {self.b} x + {self.x}>"

    def __call__(self, x):
        """ Shorthand to get y for a given x """
        if x is None:
            return None
        result = (self.a * pow(x, 2)) + (self.b * x) + self.c
        return result

    def intersect(self, q2):
        """ Get the x coordinates of the intersections of the two quadratics """
        assert(isinstance(q2, Quadratic))
        aprime = q2.a - self.a
        bprime = q2.b - self.b
        cprime = q2.c - self.c

        q3 = Quadratic(aprime, bprime, cprime)

        xs = q3.solve()
        xs_existing = np.array([x for x in xs if x is not None])
        xs_existing.sort()
        return xs_existing

    def discriminant(self):
        """ Get the discriminant of the quadratic """
        return pow(self.b, 2) - (4 * self.a * self.c)

    def solve(self):
        """ Solve the quadratic, returning up to two values"""
        return_value = []
        discriminant = self.discriminant()
        numerator_a  = -self.b
        denominator  = 2 * self.a

        if discriminant < 0:
            return_value = [None, None]
        elif np.allclose(discriminant, 0) or np.allclose(self.a, 0):
            logging.debug('Only one intersection')
            #using mullers method:
            twoc = -2 * self.c
            sqrt_discriminant = sqrt(discriminant)
            pos = (-self.b) + sqrt_discriminant
            neg = (-self.b) - sqrt_discriminant
            if pos != 0:
                x = twoc / pos
            elif neg != 0:
                x = twoc / neg
            else:
                logging.debug("Not even one intersection")
                x = None
            return_value = [x, None]
        else:
            z = sqrt(discriminant)
            return_value = [
                (numerator_a + z) / denominator,
                (numerator_a - z) / denominator,
            ]

        return return_value

    # def solve(self):
    #     twoA = 2 * self.a
    #     sqrtb4ac = sqrt(pow(self.b, 2) - (4 * self.a * self.c))
    #     pos = -self.b + sqrtb4ac
    #     neg = -self.b - sqrtb4ac
    #     return [neg, pos]



"""
Utility module to assist with drawing and debugging voronoi
"""
##-- imports
from __future__ import annotations
import logging as root_logger
from math import nan
from os.path import isdir, join
import numpy as np

##-- end imports

logging = root_logger.getLogger(__name__)

#Constants:
COLOUR                    = [0.2, 0.1, 0.6, 1.0]
COLOUR_TWO                = [1.0, 0.2, 0.4, 0.5]
SITE_COLOUR               = [1, 0, 0, 1]
SITE_RADIUS               = 10
CIRCLE_COLOUR             = [1, 1, 0, 1]
CIRCLE_COLOUR_INACTIVE    = [0, 0, 1, 1]
CIRCLE_RADIUS             = 10
BEACH_LINE_COLOUR         = [0, 1, 0]
BEACH_LINE_COLOUR_TWO     = [1, 1, 0]
BEACH_NO_INTERSECT_COLOUR = [1, 0, 0, 1]
BEACH_RADIUS              = 10
SWEEP_LINE_COLOUR         = [0, 0, 1, 1]
LINE_WIDTH                = 3
NUM_POINTS                = 2000

class VoronoiDebug:
    """ Contains methods to draw various stages of the voronoi calculation """

    def __init__(self, n, directory, instance):
        surf, ctx, size, n = utils.drawing.setup_cairo(n=n, scale=False, cartesian=True)
        assert(isdir(directory))
        self.instance = instance
        self.surface = surf
        self.ctx = ctx
        self.draw_size = size
        self.draw_n = n
        self.save_dir = directory
        self.bbox = self.instance.bbox

    def draw_voronoi_diagram(self,
                             i=None,
                             clear=True,
                             faces=True,
                             edges=False,
                             verts=False,
                             text=False):
        """ Draw the final diagram """
        logging.debug("------------------------------")
        logging.debug("Drawing final voronoi diagram")
        dcel = self.instance.finalise_DCEL()
        if clear:
            utils.drawing.clear_canvas(self.ctx, bbox=self.bbox)
        self.ctx.set_source_rgba(*COLOUR)
        #draw sites
        for site in self.instance.sites:
            utils.drawing.draw_circle(self.ctx, *site.loc, 0.007)
        #draw faces
        utils.dcel.dcel_drawing.draw_dcel(self.ctx, dcel,
                                          text=text,
                                          faces=faces, edges=edges, verts=verts)
        utils.drawing.write_to_png(self.surface, join(self.save_dir, "voronoi_debug"))

    def draw_intermediate_states(self, i,
                                 sites=True,
                                 beachline=True,
                                 sweepline=True,
                                 circles=True,
                                 dcel=False,
                                 text=False,
                                 face=None,
                                 ind_face=False):
        """ Top level function to draw intermediate state of the algorithm """
        logging.info("Drawing intermediate state: {}".format(i))
        if ind_face:
            for f in self.instance.dcel.faces:
                self.draw_intermediate_states(f.index,
                                              face=f,
                                              text=True,
                                              sites=sites,
                                              beachline=beachline,
                                              sweepline=sweepline,
                                              circles=circles)

        utils.drawing.clear_canvas(self.ctx, bbox=self.bbox)

        if sites:
            self.draw_sites()
        if beachline:
            self.draw_beach_line_components()
        if sweepline:
            self.draw_sweep_line()
        if circles:
            self.draw_circle_events()
        if dcel:
            #TODO: draw the incomplete lines better
            #backup_dcel_data = self.instance.dcel.export_data()
            #dcelInstance = self.instance.finalise_DCEL()
            utils.dcel.dcel_drawing.draw_dcel(self.ctx, self.instance.dcel,
                                              background_colour=[0, 0, 0, 0],
                                              faces=False,
                                              edges=True,
                                              verts=True,
                                              text=text)
        if face is not None:
            face.draw(self.ctx, clear=False)

        #self.instance.dcel.import_data(backup_dcel_data)
        utils.drawing.write_to_png(self.surface, join(self.save_dir, "voronoi_intermediate"), i=i)

    def draw_sites(self):
        """ Draw the individual sites """
        self.ctx.set_source_rgba(*SITE_COLOUR)
        for site in self.instance.sites:
            utils.drawing.draw_circle(self.ctx, *site.loc, SITE_RADIUS)

    def draw_circle_events(self):
        """ Draw the circle events """
        for event in self.instance.circles:
            if event.active:
                self.ctx.set_source_rgba(*CIRCLE_COLOUR)
                utils.drawing.draw_circle(self.ctx, *event.loc, CIRCLE_RADIUS)
            else:
                self.ctx.set_source_rgba(*CIRCLE_COLOUR_INACTIVE)
                utils.drawing.draw_circle(self.ctx, *event.loc, CIRCLE_RADIUS)

    def draw_beach_line_components(self):
        """ Draw the arcs of the beach line """
        #the arcs themselves
        self.ctx.set_source_rgba(*BEACH_LINE_COLOUR, 0.1)
        xs = np.linspace(self.bbox[0], self.bbox[2], NUM_POINTS)
        for node in self.instance.beachline.nodes:
            arc = node.value
            xys = arc(xs)
            for x, y in xys:
                utils.drawing.draw_circle(self.ctx, x, y, BEACH_RADIUS)
        #--------------------
        #the frontier:
        # Essentially a horizontal travelling sweep line to draw segments
        self.ctx.set_source_rgba(*BEACH_LINE_COLOUR_TWO, 1)
        leftmost_x = nan
        ##Get the chain of arcs:
        chain = self.instance.beachline.get_chain()
        if len(chain) > 1:
            enumerated = list(enumerate(chain))
            logging.debug("Chain: {}".format("".join([str(x) for x in chain])))
            pairs = zip(enumerated[0:-1], enumerated[1:])
            for (i, a), (j, b) in pairs:
                logging.debug("Drawing {} -> {}".format(a, b))
                intersections = a.value.intersect(b.value, self.instance.sweep_position.y())
                logging.debug("Intersections: {}".format(intersections))
                if not bool(intersections):
                    logging.exception("NO INTERSECTION: {} - {}".format(i, j))
                    #Draw the non-intersecting line as red
                    self.ctx.set_source_rgba(*BEACH_NO_INTERSECT_COLOUR)
                    xs = np.linspace(self.bbox[0], self.bbox[2], NUM_POINTS)
                    axyrs = np.column_stack((a.value(xs), BEACH_RADIUS))
                    bxyrs = np.column_stack((b.value(xs), BEACH_RADIUS))

                    utils.drawing.draw_circle(self.ctx, axyrs)
                    utils.drawing.draw_circle(self.ctx, bxyrs)
                    self.ctx.set_source_rgba(*BEACH_LINE_COLOUR_TWO, 1)
                    continue
                #----------
                #intersection xs:
                i_xs = intersections[:, 0]
                #xs that are further right than what we've drawn
                if leftmost_x is nan:
                    valid_xs = i_xs
                else:
                    valid_xs = i_xs[leftmost_x < i_xs]
                if not bool(valid_xs):
                    #nothing valid, try the rest of the arcs
                    logging.debug("No valid xs, continuing")
                    continue
                left_most_intersection = valid_xs.min()
                logging.debug("Arc {0} from {1:.2f} to {2:.2f}".format(i,
                                                                       leftmost_x,
                                                                       left_most_intersection))
                if leftmost_x is nan:
                    leftmost_x = left_most_intersection - 1
                xs = np.linspace(leftmost_x, left_most_intersection, NUM_POINTS)
                #update the position
                leftmost_x = left_most_intersection
                frontier_arc = np.column_stack((a.value(xs), BEACH_RADIUS))
                utils.drawing.draw_circle(self.ctx, frontier_arc)

        if bool(chain) and (leftmost_x is nan or leftmost_x < self.bbox[2]):
            if leftmost_x is nan:
                leftmost_x = 0
            #draw the last arc:
            logging.debug("Final Arc: {}".format(str(chain[-1])))
            logging.debug("Final Arc from {0:.2f} to {1:.2f}".format(leftmost_x, self.bbox[2]))
            xs = np.linspace(leftmost_x, self.bbox[2], NUM_POINTS)
            frontier_arc = np.column_stack((chain[-1].value(xs), BEACH_RADIUS))
            utils.drawing.draw_circle(self.ctx, frontier_arc)

    def draw_sweep_line(self):
        """ Draw the horizontal sweep line """
        if self.instance.sweep_position is None:
            return
        self.ctx.set_source_rgba(*SWEEP_LINE_COLOUR)
        self.ctx.set_line_width(LINE_WIDTH)
        #a tuple
        sweep_event = self.instance.sweep_position
        self.ctx.move_to(self.bbox[0], sweep_event.y())
        self.ctx.line_to(self.bbox[2], sweep_event.y())
        self.ctx.close_path()
        self.ctx.stroke()



#!/usr/bin/env python3
# from https://alexandra-zaharia.github.io/posts/make-your-own-custom-color-formatter-with-python-logging/
##-- imports
from __future__ import annotations

import logging
import warnings
import re
from collections import defaultdict
from string import Formatter
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

import _string
##-- end imports

##-- colours import attempt
try:
    # If `sty` is installed, use that
    from sty import bg, ef, fg, rs
    LEVEL_MAP    = defaultdict(lambda: rs.all)
    COLOUR_RESET = rs.all
    LEVEL_MAP.update({
        logging.DEBUG    : fg.grey,
        logging.INFO     : fg.blue,
        logging.WARNING  : fg.yellow,
        logging.ERROR    : fg.red,
        logging.CRITICAL : fg.red,
        "blue"           : fg.blue,
        "cyan"           : fg.cyan,
        "green"          : fg.green,
        "magenta"        : fg.magenta,
        "red"            : fg.red,
        "yellow"         : fg.yellow,
        "bg_blue"        : bg.blue,
        "bg_cyan"        : bg.cyan,
        "bg_green"       : bg.green,
        "bg_magenta"     : bg.magenta,
        "bg_red"         : bg.red,
        "bg_yellow"      : bg.yellow,
        "bold"           : ef.bold,
        "underline"      : ef.u,
        "italic"         : ef.italic,
        "RESET"          : rs.all
        })
except ImportError:
    # Otherwise don't add colours
    LEVEL_MAP    = defaultdict(lambda: "")
    COLOUR_RESET = ""
##-- end colours import attempt

class SimpleLogColour:
    """ Utility class for wrapping strings with specific colours """

    def __init__(self):
        raise TypeError("SimpleLogColour is Static, don't instance it")

    @staticmethod
    def green(s):
        return LEVEL_MAP['green'] + str(s) + COLOUR_RESET

    @staticmethod
    def blue(s):
        return LEVEL_MAP['cyan'] + str(s) + COLOUR_RESET

    @staticmethod
    def yellow(s):
        return LEVEL_MAP['yellow'] + str(s) + COLOUR_RESET

    @staticmethod
    def red(s):
        return LEVEL_MAP['red'] + str(s) + COLOUR_RESET

class JGDVColourFormatter(logging.Formatter):
    """
    Stream Formatter for doot, enables use of colour sent to console

    Guarded Formatter for adding colour.
    Uses the sty module.
    If sty is missing, behaves as the default formatter class

    # Do *not* use for on filehandler
    Usage reminder:
    # Create stdout handler for logging to the console (logs all five levels)
    stdout_handler = logging.StreamHandler()
    stdout_handler.setFormatter(DootLogFormatter(fmt))
    logger.addHandler(stdout_handler)
    """

    _default_fmt : ClassVar[str] = '{asctime} | {levelname:9} | {message}'
    _default_date_fmt : str      =  "%H:%M:%S"
    _default_style               = '{'

    def __init__(self, *, fmt=None):
        """
        Create the DootLogFormatter with a given *Brace* style log format
        """
        super().__init__(fmt or self._default_fmt,
                         datefmt=self._default_date_fmt,
                         style=self._default_style)
        self.colours = LEVEL_MAP

    def format(self, record):
        log_colour = self.colours[record.levelno]
        if hasattr(record, "colour"):
            log_colour = self.colours[record.colour]

        return log_colour + super().format(record) + COLOUR_RESET

class JGDVColourStripFormatter(logging.Formatter):
    """
    Force Colour Command codes to be stripped out of a string.
    Useful for when you redirect printed strings with colour
    to a file
    """

    _default_fmt         = "{asctime} | {levelname:9} | {shortname:25} | {message}"
    _default_date_fmt    = "%Y-%m-%d %H:%M:%S"
    _default_style       = '{'
    _colour_strip_re     = re.compile(r'\x1b\[([\d;]+)m?')

    def __init__(self, *, fmt=None):
        """
        Create the DootLogFormatter with a given *Brace* style log format
        `record` will install the DootLogRecord as the record factory if true
        """
        super().__init__(fmt or self._default_fmt,
                         datefmt=self._default_date_fmt,
                         style=self._default_style)

    def format(self, record):
        result    = super().format(record)
        no_colour = self._colour_strip_re.sub("", result)
        return no_colour



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

TAG_NORM : Final[re.Pattern] = re.compile(" +")

@dataclass
class Bookmark:
    url     : str      = field()
    tags    : Set[str] = field(default_factory=set)
    name    : str      = field(default="No Name")
    sep     : str      = field(default=" : ")

    @staticmethod
    def build(line:str, sep=None):
        """
        Build a bookmark from a line of a bookmark file
        """
        sep  = sep or Bookmark.sep
        tags = []
        match [x.strip() for x in line.split(sep)]:
            case []:
                raise TypeException("Bad line passed to Bookmark")
            case [url]:
                logging.warning("No Tags for: %s", url)
            case [url, *tags]:
                pass

        return Bookmark(url,
                        set(tags),
                        sep=sep)

    def __post_init__(self):
        self.tags = {TAG_NORM.sub("_", x.strip()) for x in self.tags}

    def __eq__(self, other):
        return self.url == other.url

    def __lt__(self, other):
        return self.url < other.url

    def __str__(self):
        tags = self.sep.join(sorted(self.tags))
        return f"{self.url}{self.sep}{tags}"

    @property
    def url_comps(self) -> url_parse.ParseResult:
        return url_parse.urlparse(self.url)

    def merge(self, other) -> 'Bookmark':
        """ Merge two bookmarks' tags together,
        creating a new bookmark
        """
        assert(self == other)
        merged = Bookmark(self.url,
                          self.tags.union(other.tags),
                          self.name,
                          sep=self.sep)
        return merged

    def clean(self, subs):
        """
        run tag substitutions on all tags in the bookmark
        """
        cleaned_tags = set()
        for tag in self.tags:
            cleaned_tags.add(subs.sub(tag))

        self.tags = cleaned_tags



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class JGDVNonKey(str, JGDVBaseKey):
    """
      Just a string, not a key. But this lets you call no-ops for key specific methods
    """

    def __repr__(self):
        return "<DootNonKey: {}>".format(str(self))

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        match other:
            case JGDVBaseKey() | str():
                return str(self) == str(other)
            case _:
                return False

    def within(self, other:str|dict|TomlGuard) -> bool:
        match other:
            case str():
                return self.form in other
            case dict() | TomlGuard():
                return self in other
            case _:
                raise TypeError("Uknown JGDVBaseKey target for within", other)

    @property
    def indirect(self) -> JGDVBaseKey:
        if not self.is_indirect:
            return DootSimpleKey("{}_".format(super().__str__()))
        return self

    @property
    def is_indirect(self):
        return False

    @property
    def form(self):
        """ Return the key in its use form """
        return str(self)

    def expand(self, spec=None, state=None, *, rec=False, insist=False, chain:list[JGDVBaseKey]=None, on_fail=Any, locs:DootLocations=None, **kwargs) -> str:
        return str(self)

    def redirect(self, spec=None) -> JGDVBaseKey:
        return self

    def to_type(self, spec, state, type_=Any, **kwargs) -> str:
        if type_ not in [Any, str]:
            raise TypeError("NonKey's can only be strings", self, type_)
        return str(self)

class JGDVSimpleKey(str, JGDVBaseKey):
    """
      A Single key with no extras.
      ie: {x}. not {x}{y}, or {x}.blah.
    """

    def __repr__(self):
        return "<DootSimpleKey: {}>".format(str(self))

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        match other:
            case JGDVBaseKey() | str():
                return str(self) == str(other)
            case _:
                return False

    @property
    def indirect(self):
        if not self.is_indirect:
            return DootSimpleKey("{}_".format(super().__str__()))
        return self

    @property
    def is_indirect(self):
        return str(self).endswith("_")

    @property
    def form(self):
        """ Return the key in its use form, ie: wrapped in braces """
        return "{{{}}}".format(str(self))

    def within(self, other:str|dict|TomlGuard) -> bool:
        match other:
            case str():
                return self.form in other
            case dict() | TomlGuard():
                return self in other
            case _:
                raise TypeError("Uknown JGDVBaseKey target for within", other)

    def expand(self, spec=None, state=None, *, rec=False, insist=False, chain:list[JGDVBaseKey]=None, on_fail=Any, locs:DootLocations=None, **kwargs) -> str:
        key = self.redirect(spec)
        try:
            return DootFormatter.fmt(key, _spec=spec, _state=state, _rec=rec, _locs=locs, _insist=insist)
        except (KeyError, TypeError) as err:
            if bool(chain):
                return chain[0].expand(spec, state, rec=rec, chain=chain[1:], on_fail=on_fail)
            elif on_fail != Any:
                return on_fail
            else:
                raise err

    def redirect(self, spec:None|SpecStruct_p=None) -> JGDVBaseKey:
        """
          If the indirect form of the key is found in the spec, use that as a key instead
        """
        if not spec:
            return self

        match spec:
            case SpecStruct_p():
                kwargs = spec.params
            case None:
                kwargs = {}

        match kwargs.get(self.indirect, self):
            case str() as x if x == self.indirect:
                return self
            case str() as x:
                return JGDVBaseKey.build(x)
            case list() as lst:
                raise TypeError("Key Redirection resulted in a list, use redirect_multi", self)

        return self

    def redirect_multi(self, spec:None|SpecStruct_p=None) -> list[JGDVBaseKey]:
        """ redirect an indirect key to a *list* of keys """
        if not spec:
            return [self]

        match spec:
            case SpecStruct_p():
                kwargs = spec.params
            case None:
                kwargs = {}

        match kwargs.get(self.indirect, self):
            case str() as x if x == self:
                return [self]
            case str() as x:
                return [JGDVBaseKey.build(x)]
            case list() as lst:
                return [JGDVBaseKey.build(x) for x in lst]

        return [self]

    def to_type(self, spec:None|SpecStruct_p=None, state=None, type_=Any, chain:list[JGDVBaseKey]=None, on_fail=Any) -> Any:
        target            = self.redirect(spec)

        match spec:
            case SpecStruct_p():
                kwargs = spec.params
            case None:
                kwargs = {}

        task_name = state.get(STATE_TASK_NAME_K, None) if state else None
        match (replacement:=DootKeyGetter.chained_get(target, kwargs, state)):
            case None if bool(chain):
                return chain[0].to_type(spec, state, type_=type_, chain=chain[1:], on_fail=on_fail)
            case None if on_fail != Any and isinstance(on_fail, JGDVBaseKey):
                return on_fail.to_type(spec, state, type_=type_)
            case None if on_fail != Any:
                return on_fail
            case None if type_ is Any or type_ is None:
                return None
            case _ if type_ is Any:
                return replacement
            case _ if type_ and isinstance(replacement, type_):
                return replacement
            case None if not any(target in x for x in [kwargs, state]):
                raise KeyError("Key is not available in the state or spec", target)
            case _:
                raise TypeError("Unexpected Type for replacement", type_, replacement, self)

class JGDVArgsKey(str, JGDVBaseKey):
    """ A Key representing the action spec's args """

    def __call__(self, spec, state, **kwargs):
        return self.to_type(spec, state)

    def __repr__(self):
        return "<DootArgsKey>"

    def expand(self, *args, **kwargs):
        raise doot.errors.DootKeyError("Args Key doesn't expand")

    def redirect(self, spec=None):
        raise doot.errors.DootKeyError("Args Key doesn't redirect")

    def to_type(self, spec=None, state=None, *args, **kwargs) -> list:
        return spec.args

class JGDVKwargsKey(JGDVArgsKey):
    """ A Key representing all of an action spec's kwargs """

    def __repr__(self):
        return "<DootArgsKey>"

    def to_type(self, spec:None|SpecStruct_p=None, state=None, *args, **kwargs) -> dict:
        match spec:
            case SpecStruct_p():
                return spec.params
            case None:
                return {}

class JGDVImportKey(JGDVSimpleKey):
    """ a key to specify a key is used for importing
    ie: str expands -> DootCodeReference.build -> .try_import
    """
    pass



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import datetime
import json
import logging as logmod
import pathlib as pl
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
from uuid import uuid1

import networkx as nx
##-- end imports

logging = logmod.getLogger(__name__)

@dataclass
class TwitterThreadObj:
    """ Utility Class to hold a thread
    doesn't hold the tweets themselves, just ID's
    """
    main      : list[str]       = field(default_factory=list)
    rest      : list[list[str]] = field(default_factory=list)
    quotes    : list[str]       = field(default_factory=list)
    component : pl.Path         = field(default=None)
    base_user : str             = field(default=None)

    @staticmethod
    def build(data):
        assert(all([x in data for x in ["main_thread",
                                        "rest",
                                        "quotes"]]))

        return TwitterThreadObj(data["main_thread"],
                                data["rest"],
                                data["quotes"])

    def dump(self):
        return {
            'main_thread' : self.main,
            'rest'        : self.rest,
            'quotes'      : self.quotes,
            'component'   : self.component,
            'base_user'   : self.base_user,
            }




#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##-- imports
from importlib.resources import files
##-- end imports

import tomlguard
from jgdv.spiders.crawler import CrawlerProcessFix
from urllib.parse import urlparse

default_toml           = tomlguard.load(files("doot.__templates") / "spider_toml").flatten_on().spiders().get_table()
spider_settings        = doot.config.flatten_on().spiders().get_table()

settings_with_defaults = default_toml.copy()
settings_with_defaults.update(spider_settings)

class Spider_M:
    """
    Run a scrapy spider either with default doot settings,
    or passed in toml data
    """

    def _urlparse(self, url:str):
        return urlparse(url)

    def run_spider(self, name:str, spider:type, urls:list, settings=None, auto_limit=True, scrapy_log=False):
        """
        auto_limit = True -> auto set allowed_domains to the domains of initial urls
        """
        logging.info("Running spider: %s : %s : %s : %s", name, spider, urls, settings)
        merged = settings_with_defaults
        match settings:
            case None:
                pass
            case dict() if bool(settings):
                merged.update(settings)
            case tomlguard.TomlGuard():
                merged.update(dict(settings))
            case [dict() as val]:
                merged = val

        allowed_domains = [self._urlparse(x).netloc for x in urls] if auto_limit else []
        logging.debug("----- Building Crawler")
        self.crawler    = CrawlerProcessFix(settings=merged, install_root_handler=scrapy_log)
        self.crawler.crawl(spider,
                           name=name,
                           locs=self.locs,
                           urls=urls,
                           domains=allowed_domains,
                           task=self)
        logging.debug("------ Starting Crawl")
        self.crawler.start()
        logging.debug("------ Crawl Finished")




#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

# import abc
# import datetime
# import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import importlib
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
# If CLI:
# logging = logmod.root
# logging.setLevel(logmod.NOTSET)
##-- end logging

from collections import defaultdict
from importlib.metadata import entry_points, EntryPoint
import tomlguard
import jgdv
from jgdv._interfaces.loader import PluginLoader_p

skip_default_plugins        = doot.config.on_fail(False).skip_default_plugins()
skip_plugin_search          = doot.config.on_fail(False).skip_plugin_search()
env_plugins                 = doot.config.on_fail({}).plugins(wrapper=dict)
plugin_types                = set(doot.constants.entrypoints.FRONTEND_PLUGIN_TYPES + doot.constants.entrypoints.BACKEND_PLUGIN_TYPES)
cmd_loader_key  : Final     = doot.constants.entrypoints.DEFAULT_COMMAND_LOADER_KEY
task_loader_key : Final     = doot.constants.entrypoints.DEFAULT_TASK_LOADER_KEY

def make_ep (x, y, z):
    if z not in plugin_types:
        raise doot.errors.DootPluginError("Plugin Type Not Found: %s : %s", z, (x, y))
    return EntryPoint(name=x, value=y, group="{}.{}".format(doot.constants.entrypoints.PLUGIN_TOML_PREFIX, z))


@jgdv.check_protocol
class JGDVPluginLoader(PluginLoader_p):
    """
    Load doot plugins from the system, to choose from with doot.toml or cli args
    """

    def setup(self, extra_config=None) -> Self:
        self.plugins = defaultdict(list)
        match extra_config:
            case None:
                self.extra_config = tomlguard.TomlGuard({})
            case dict():
                self.extra_config = tomlguard.TomlGuard(extra_config)
            case tomlguard.TomlGuard():
                self.extra_config = extra_config

        return self

    def load(self) -> TomlGuard[EntryPoint]:
        """
        use entry_points(group="doot")
        add to the config tomlguard
        """
        logging.debug("---- Loading Plugins: %s", doot.constants.entrypoints.PLUGIN_TOML_PREFIX)
        try:
            self._load_system_plugins()
        except Exception as err:
            raise doot.errors.DootPluginError("Failed to load system wide plugins: %s", err) from err

        try:
            self._load_from_toml()
        except Exception as err:
            raise doot.errors.DootPluginError("Failed to load toml specified plugins: %s", err) from err

        try:
            self._load_extra_plugins()
        except Exception as err:
            raise doot.errors.DootPluginError("Failed to load command line/dooter specified plugins: %s", err) from err

        try:
            self._append_defaults()
        except Exception as err:
            raise doot.errors.DootPluginError("Failed to load plugin defaults: %s", err) from err

        logging.debug("Found %s plugins", len(self.plugins))
        PluginLoader_p.loaded = tomlguard.TomlGuard(self.plugins)
        return PluginLoader_p.loaded

    def _load_system_plugins(self):
        if skip_plugin_search:
            pass
        else:
            logging.info("Searching environment for plugins, skip with `skip_plugin_search` in config")
            for plugin_type in plugin_types:
                try:
                    plugin_group = "{}.{}".format(doot.constants.entrypoints.PLUGIN_TOML_PREFIX, plugin_type)
                    # Load env wide entry points
                    for entry_point in entry_points(group=plugin_group):
                        self.plugins[plugin_type].append(entry_point)
                except Exception as err:
                    raise doot.errors.DootPluginError("Plugin Failed to Load: %s : %s : %s", plugin_group, entry_point, err) from err

    def _load_from_toml(self):
        # load config entry points
        for cmd_group, vals in env_plugins.items():
            if cmd_group not in plugin_types:
                logging.warning("Unknown plugin type found in config: %s", cmd_group)
                continue
            if not isinstance(vals, (tomlguard.TomlGuard, dict)):
                logging.warning("Toml specified plugins need to be a dict of (cmdName : class): %s ", cmd_group)
                continue

            for name, cls in vals.items():
                ep = EntryPoint(name=name, value=cls, group=cmd_group)
                self.plugins[cmd_group].append(ep)

    def _load_extra_plugins(self):
        extra_eps    = self.extra_config.on_fail({}).plugins(wrapper=dict)
        # load extra-config entry points
        for k,v in extra_eps.items():
            if k not in plugin_types:
                logging.warning("Unknown plugin type found in extra config: %s", k)
                continue
            ep = EntryPoint(name=k, value=v, group=doot.constants.entrypoints.PLUGIN_TOML_PREFIX)
            logging.debug("Adding Plugin: %s", ep)
            self.plugins[k].append(ep)

    def _append_defaults(self):
        if skip_default_plugins:
            logging.info("Skipping Default Plugins")
            return

        self.plugins[cmd_loader_key].append(make_ep(cmd_loader_key, "doot.loaders.cmd_loader:DootCommandLoader", cmd_loader_key))
        self.plugins[task_loader_key].append(make_ep(task_loader_key, "doot.loaders.task_loader:DootTaskLoader", task_loader_key))

        for group, vals in doot.aliases:
            logging.debug("Loading aliases: %s", group)
            self.plugins[group]  += [make_ep(x, y, group) for x,y in vals.items()]



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class BookmarkCollection:

    entries : List[Bookmark] = field(default_factory=list)
    ext     : str            = field(default=".bookmarks")

    @staticmethod
    def read(fpath:pl.Path) -> BookmarkCollection:
        """ Read a file to build a bookmark collection """
        bookmarks = BookmarkCollection()
        for line in (x.strip() for x in fpath.read_text().split("\n")):
            if not bool(line):
                continue
            bookmarks += Bookmark.build(line)

        return bookmarks

    def __str__(self):
        return "\n".join(map(str, sorted(self.entries)))

    def __repr__(self):
        return f"<{self.__class__.__name__}: {len(self)}>"

    def __iadd__(self, value):
        return self.update(value)

    def __iter__(self):
        return iter(self.entries)

    def __contains__(self, value:Bookmark):
        return value in self.entries

    def __len__(self):
        return len(self.entries)

    def __hash__(self):
        return id(self)
    def update(self, *values):
        for val in values:
            match val:
                case Bookmark():
                    self.entries.append(val)
                case BookmarkCollection():
                    self.entries += val.entries
                case [*vals] | set(vals):
                    self.update(*vals)
                case _:
                    raise TypeError(type(val))
        return self

    def difference(self, other:BookmarkCollection):
        result = BookmarkCollection()
        for bkmk in other:
            if bkmk not in self:
                result += bkmk

        return result

    def merge_duplicates(self):
        deduplicated = {}
        for x in self:
            if x.url not in deduplicated:
                deduplicated[x.url] = x
            else:
                deduplicated[x.url] = x.merge(deduplicated[x.url])

        self.entries = list(deduplicated.values())



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

# import abc
# import datetime
# import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
# from uuid import UUID, uuid1
# from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import sys

class FrameHelper:
    """
      Utility to work with python frames and tracebacks
    """

    def __init__(self):
        self.frames = []
        self.get_frames()

    def __getitem__(self, val=None):
        match val:
            case None:
                return self.to_tb()
            case slice() | int():
                return self.to_tb(self.frames[val])
            case _:
                raise TypeError("Bad value passed to TraceHelper")

    def get_frames(self):
        """ from https://stackoverflow.com/questions/27138440 """
        tb    = None
        depth = 0
        while True:
            try:
                frame = sys._getframe(depth)
                depth += 1
            except ValueError as exc:
                break

            self.frames.append(frame)

    def to_tb(self, frames=None):
        top = None
        frames = frames or self.frames
        for frame in frames:
            top = types.TracebackType(top, frame,
                                     frame.f_lasti,
                                     frame.f_lineno)
        return top



"""
How to convert a (non-human friendly) model trace into a human readable format
"""
##-- imports
from __future__ import annotations

import abc
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from string import Template
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

from jgdv.apis.clingo.trace import Trace_i

@dataclass
class SolverReporter_i(metaclass=abc.ABCMeta):
    """
        Solver Report Generator interface
    """

    def __init__(self):
        self._compiled_text : list[str] = []

    def clear(self):
        self._compiled_text = []

    def expand(self, pattern:str|Template, **kwargs):
        match pattern:
            case Template():
                return pattern.safe_substitute(kwargs)
            case str() if not bool(kwargs):
                return pattern
            case str():
                return pattern.format_map(kwargs)
            case _:
                raise TypeError("Unrecognised compile pattern type", pattern)

    def insert(self, pattern:str|Template, **kwargs):
        """
        insert a given pattern text into the compiled output,
        formatting it with kwargs.
        """
        self._compiled_text.append(self.expand(pattern, **kwargs))

    @abc.abstractmethod
    def trace_to_file(self, trace:Trace_i, path:Path): pass



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import importlib
from importlib.metadata import EntryPoint
from tomlguard import TomlGuard
from jgdv.structs.name import StructuredName

@dataclass(eq=False, slots=True)
class JGDVCodeReference(StructuredName):
    """
      A reference to a class or function. can be created from a string (so can be used from toml),
      or from the actual object (from in python)
    """
    separator : str                              = field(default=doot.constants.patterns.IMPORT_SEP, kw_only=True)
    _mixins   : list[JGDVCodeReference]          = field(default_factory=list, kw_only=True)
    _type     : None|type                        = field(default=None, kw_only=True)

    @classmethod
    def build(cls, name:str|type|EntryPoint):
        match name:
            case str():
                return cls._from_str(name)
            case type():
                groupHead = name.__module__
                codeHead  = name.__name__
                ref = cls(groupHead, codeHead, _type=name)
                return ref
            case EntryPoint():
                loaded = ctor.load()
                groupHead = loaded.__module__
                codeHead  = loaded.__name__
                ref = cls(groupHead, codeHead, _type=loaded)
                return ref

    @classmethod
    def _from_str(cls, name:str):
        if doot.constants.patterns.TASK_SEP in name:
            raise doot.errors.DootError("Code References should use a single colon, not double")

        if ":" in name:
            try:
                groupHead_r, taskHead_r = name.split(":")
                groupHead = groupHead_r.split(".")
                taskHead = taskHead_r.split(".")
            except ValueError:
                raise doot.errors.DootConfigError("Code ref can't be split correctly, ensure its of the form x.y.z:ClassName", name)
        else:
            groupHead = None
            taskHead  = name

        return JGDVCodeReference(groupHead, taskHead)

    @staticmethod
    def from_alias(alias:str, group:str, plugins:TomlGuard) -> JGDVCodeReference:
        if group not in plugins:
            return JGDVCodeReference._from_str(alias)
        match [x for x in plugins[group] if x.name == alias]:
            case [x, *xs]:
                return JGDVCodeReference._from_str(x.value)
            case _:
                return JGDVCodeReference._from_str(alias)

    def __str__(self) -> str:
        return "{}{}{}".format(self.module, self.separator, self.value)

    def __repr__(self) -> str:
        code_path = str(self)
        mixins    = ", ".join(str(x) for x in self._mixins)
        return f"<CodeRef: {code_path} Mixins: {mixins}>"

    def __hash__(self):
        return hash(str(self))

    def __iter__(self):
        return iter(self._mixins)

    @property
    def module(self):
        return self.subseparator.join(self.head)

    @property
    def value(self):
        return self.subseparator.join(self.tail)

    def add_mixins(self, *mixins:str|JGDVCodeReference|type, plugins:TomlGuard=None) -> JGDVCodeReference:
        to_add = self._mixins[:]
        for mix in mixins:
            match mix:
                case JGDVCodeReference():
                    ref = mix
                case str() if plugins is not None:
                    ref = JGDVCodeReference.from_alias(mix, "mixin", plugins)
                case str() | type():
                    ref = JGDVCodeReference.build(mix)
                case _:
                    raise TypeError("Unrecognised mixin type", mix)

            if ref not in to_add:
                to_add.append(ref)

        new_ref = JGDVCodeReference(head=self.head[:], tail=self.tail[:], _mixins=to_add, _type=self._type)
        return new_ref

    def try_import(self, ensure:type=Any) -> Any:
        try:
            if self._type is not None:
                curr = self._type
            else:
                mod = importlib.import_module(self.module)
                curr = mod
                for name in self.tail:
                    curr = getattr(curr, name)

            if bool(self._mixins):
                mixins = []
                for mix in self._mixins:
                    match mix:
                        case JGDVCodeReference():
                            mixins.append(mix.try_import())
                        case type():
                            mixins.append(mix)
                curr = type(f"DootGenerated:{curr.__name__}", tuple(mixins + [curr]), {})

            if ensure is not Any and not (isinstance(curr, ensure) or issubclass(curr, ensure)):
                raise ImportError("Imported Code Reference is not of correct type", self, ensure)

            return curr
        except ModuleNotFoundError as err:
            raise ImportError("Module can't be found", str(self))
        except AttributeError as err:
            raise ImportError("Attempted to import %s but failed", str(self)) from err

    def to_aliases(self, group:str, plugins:TomlGuard) -> tuple[str, list[str]]:
        base_alias = str(self)
        match [x for x in plugins[group] if x.value == base_alias]:
            case [x, *xs]:
                base_alias = x.name

        if group != "mixins":
            mixins = [x.to_aliases("mixins", plugins)[0] for x in  self._calculate_minimal_mixins(plugins)]
        else:
            mixins = []

        return base_alias, mixins

    def _calculate_minimal_mixins(self, plugins:TomlGuard) -> list[JGDVCodeReference]:
        found          = set()
        minimal_mixins = []
        for mixin in reversed(sorted(map(lambda x: x.try_import(), self), key=lambda x:  len(x.mro()))):
            if mixin in found:
                continue

            found.update(mixin.mro())
            minimal_mixins.append(mixin)

        return [JGDVCodeReference.build(x) for x in minimal_mixins]



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import heapq

class JGDVHeap:

    def __init__(self):
        self._heap = None

    def pop_while_same(self):
        """ Pop while the head is equal to the first value poppped """
        assert(all([isinstance(x, HeapWrapper) for x in self._heap]))
        first_vert, first_edge = heapq.heappop(self._heap).unwrap()
        if first_edge is None:
            return (first_vert, [])

        collected = (first_vert, [first_edge])
        count = 1
        while bool(self._heap) and self._heap[0].ordinal == first_vert:
            data = heapq.heappop(self._heap).data
            if data is not None:
                collected[1].append(data)
                count += 1
        assert(len(collected[1]) == count)
        return collected



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class JGDVFormatter(string.Formatter):
    """
      A Formatter for expanding arguments based on action spec kwargs, and task state, and cli args
    """
    _fmt                = None

    SPEC   : Final[str] = "_spec"
    INSIST : Final[str] = "_insist"
    STATE  : Final[str] = "_state"
    LOCS   : Final[str] = "_locs"
    REC    : Final[str] = "_rec"

    @staticmethod
    def fmt(fmt:str|DootKey|pl.Path, /, *args, **kwargs) -> str:
        if not DootFormatter._fmt:
            DootFormatter._fmt = DootFormatter()

        return DootFormatter._fmt.format(fmt, *args, **kwargs)

    def format(self, fmt:str|DootKey|pl.Path, /, *args, **kwargs) -> str:
        """ expand and coerce keys """
        self._depth = 0
        match kwargs.get(self.SPEC, None):
            case None:
                kwargs['_spec'] = {}
            case SpecStruct_p():
                kwargs['_spec'] = kwargs[self.SPEC].params
            case x:
                raise TypeError("Bad Spec Type in Format Call", x)

        match fmt:
            case DootKey():
                fmt = fmt.form
                result = self.vformat(fmt, args, kwargs)
            case str():
                result = self.vformat(fmt, args, kwargs)
            # case pl.Path():
            #     result = str(ftz.reduce(pl.Path.joinpath, [self.vformat(x, args, kwargs) for x in fmt.parts], pl.Path()))
            case _:
                raise TypeError("Unrecognized expansion type", fmt)

        return result

    def get_value(self, key, args, kwargs):
        """ lowest level handling of keys being expanded """
        logging.debug("Expanding: %s", key)
        if isinstance(key, int):
            return args[key]

        insist                = kwargs.get(self.INSIST, False)
        spec  : dict          = kwargs.get(self.SPEC, None) or {}
        state : dict          = kwargs.get(self.STATE, None) or {}
        locs  : DootLocations = kwargs.get(self.LOCS,  None)
        depth_check           = self._depth < MAX_KEY_EXPANSIONS
        rec_allowed           = kwargs.get(self.REC, False) and depth_check

        match (replacement:=DootKeyGetter.chained_get(key, spec, state, locs)):
            case None if insist:
                raise KeyError("Key Expansion Not Found")
            case None:
                return DootKey.build(key).form
            case DootKey() if depth_check:
                self._depth += 1
                return self.vformat(replacement.form, args, kwargs)
            case str() if rec_allowed:
                self._depth += 1
                return self.vformat(str(replacement), args, kwargs)
            case str():
                return replacement
            case pl.Path() if depth_check:
                self._depth += 1
                return ftz.reduce(pl.Path.joinpath, map(lambda x: self.vformat(x, args, kwargs), replacement.parts), pl.Path())
            case _:
                return str(replacement)
                # raise TypeError("Replacement Value isn't a string", args, kwargs)



"""
A Red-Black Tree Implementation
Properties of RBTrees:
1) Every node is Red Or Black
2) The root is black
3) Every leaf is Black, leaves are null nodes
4) If a node is red, it's children are black
5) All paths from a node to its leaves contain the same number of black nodes
"""
from .rbtree import RBTree
from .node import Node
from .comparison_functions import Directions



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class TwitterTweet:
    id_s         : str
    base_user    : str
    is_quote     : bool                          = field(default=False)
    name         : str                           = field(init=False, default="Unknown")
    hash_tags    : List[str]                     = field(init=False, default_factory=list)
    quote        : Tuple[str, str, TwitterTweet] = field(init=False, default=None)
    reply_to     : Tuple[str, str]               = field(init=False, default=None)
    date         : datetime                      = field(init=False, default_factory=datetime.datetime.now)
    media        : dict                          = field(default_factory=media_dict)
    links        : list[str]                     = field(default_factory=list)
    level        : int                           = field(default=4)

    fav          : int = 0
    retweet      : int = 0
    text         : str = ""

    permalink_f  : str =  "[[https://twitter.com/{}/status/{}][/{}/{}]]"



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class Node(Tree):
    """ The Container for RBTree Data """

    red     : bool     = field(default=True)
    eq_func : Callable = field(default=None)

    def __repr__(self):
        #pylint: disable=too-many-format-args
        if self.value is not None and hasattr(self.value, "id"):
            return "({}_{})".format(ascii_uppercase[self.value.id % 26],
                                    int(self.value.id/26), self.id)
        else:
            return "({}:{})".format(self.value, self.id)

    def get_black_height(self):
        """ Get the number of black nodes between self and the root """
        current = self
        height = 0
        while current is not None:
            if not current.red:
                height += 1
            current = current.parent
        return height
